import os
import torch
import shutil
from pathlib import Path
from functools import partial
from torch.nn import Module, Parameter
from torch.nn import Linear, Embedding
from torch import Tensor, LongTensor, BoolTensor
from transformers.tokenization_utils_base import EncodingFast
from typing import List, Tuple, Union, Optional, TypeVar, Iterable, Set, Dict
from torch.nn.functional import pad
from transformers import (
    PreTrainedModel,
    AutoModel,
    AutoTokenizer,
    PreTrainedTokenizer,
    BatchEncoding,
)
from .metric import CosSimilarity
from .loss import ContrastiveThresholdLoss
from ..datamodel.example import Example, strided_split, TypedSpan, collate_examples


_Model = TypeVar("_Model", bound="BI_Enc_NER")


class Base_BI_enc(Module):
    def __init__(self, param) -> None:
        super(Base_BI_enc, self).__init__()
        self._context_encoder: PreTrainedModel = AutoModel.from_pretrained(
            param.get("context_bert_model")
        )
        self._candidate_encoder: PreTrainedModel = AutoModel.from_pretrained(
            param.get("candidate_bert_model")
        )
        self._dummy_param = Parameter(torch.empty(0))

    @property
    def device(self) -> torch.device:
        return self._dummy_param.device

    def get_context_embeddings(self, context_inputs):
        return self._context_encoder(**context_inputs).last_hidden_state[:, 0]

    def get_candidate_embeddings(self, candidate_inputs):
        return self._candidate_encoder(**candidate_inputs).last_hidden_state[:, 0]

    def forward(self, context_inputs, candidate_inputs=None, **kwargs):
        cont_emb = self.get_context_embeddings(context_inputs)
        if candidate_inputs:
            cand_emb = self.get_candidate_embeddings(candidate_inputs)
            return cont_emb, cand_emb
        return cont_emb

    def save(self, save_path):
        pass

    def load(self, load_path):
        pass


class BI_Enc_NER(Base_BI_enc):
    def __init__(self, param) -> None:
        super(BI_Enc_NER, self).__init__(param)

        self._metric = CosSimilarity(scale=0.07)
        self._unk_entity_label_id = param.get("unk_entity_type_id")
        self._loss_fn_factory = partial(
            ContrastiveThresholdLoss,
            ignore_id=-100,
            reduction="mean",
            beta=param.get("loss_beta"),
        )
        self._token_encoder = self._context_encoder
        self._token_tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(
            param.get("context_bert_model")
        )
        self._descriptions_encoder = self._candidate_encoder
        self._descriptions_tokenizer: PreTrainedTokenizer = (
            AutoTokenizer.from_pretrained(param.get("candidate_bert_model"))
        )
        self._loss_fn: Optional[ContrastiveThresholdLoss] = None
        self._start_coef = param.get("start_coef")
        self._end_coef = param.get("end_coef")
        self._span_coef = param.get("span_coef")

        self._hidden_size = param.get("hidden_size")
        self._encoder_hidden = self._token_encoder.config.hidden_size
        self._max_sequence_length = min(
            param.get("max_sequence_length"), self._token_tokenizer.model_max_length
        )
        self._max_entity_length = param.get("max_entity_length")

        self._entity_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._encoded_descriptions: Optional[BatchEncoding] = None
        self._frozen_entity_representations: Optional[Tensor] = None
        self.add_descriptions(param.get("descriptions"))

        self._length_embedding = Embedding(self._max_entity_length, self._hidden_size)
        self._span_projection = Linear(
            self._encoder_hidden * 2 + self._hidden_size, self._hidden_size
        )

        self._token_start_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._token_end_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._entity_start_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._entity_end_projection = Linear(self._encoder_hidden, self._hidden_size)

    def freeze_descriptions(self):
        if self._frozen_entity_representations is not None:
            return

        with torch.no_grad():
            self._frozen_entity_representations = self._descriptions_encoder(
                self._encoded_descriptions.to(self.device)
            ).last_hidden_state[
                :, 0
            ]  # TODO: split into batches

    def drop_descriptions_encoder(self):
        self._descriptions_encoder = None

    def add_descriptions(self, descriptions: List[str]) -> None:
        self._encoded_descriptions = self._descriptions_tokenizer(
            descriptions,
            return_tensors="pt",
            truncation=True,
            max_length=self._descriptions_encoder.config.max_length,
            add_special_tokens=True,
            padding=True,
        )["input_ids"]
        self._loss_fn = self._loss_fn_factory(len(descriptions))
        self._frozen_entity_representations = None

    def eval(self: _Model) -> _Model:
        super(BI_Enc_NER, self).eval()
        self.freeze_descriptions()
        return self

    def train(self: _Model, mode: bool = True) -> _Model:
        super(BI_Enc_NER, self).train(mode)
        self._frozen_entity_representations = None
        return self

    def _get_entity_representations(self) -> Tensor:
        if self._frozen_entity_representations is not None:
            return self._frozen_entity_representations.to(self.device)
        return self._descriptions_encoder(
            self._encoded_descriptions.to(self.device)
        ).last_hidden_state[
            :, 0
        ]  # TODO: split into batches

    def _get_length_representations(self, token_representations: Tensor) -> Tensor:
        batch_size, sequence_length, representation_dims = token_representations.shape

        length_embeddings = self._length_embedding(
            torch.arange(self._max_entity_length, device=self.device)
        )
        return length_embeddings.reshape(
            1, 1, self._max_entity_length, self._hidden_size
        ).repeat(batch_size, sequence_length, 1, 1)

    def _get_span_representations(
        self, token_representations: Tensor
    ) -> Tuple[Tensor, BoolTensor]:
        batch_size, sequence_length, representation_dims = token_representations.shape

        padding_masks: List[BoolTensor] = []
        span_end_representations: List[Tensor] = []

        for shift in range(
            self._max_entity_length
        ):  # self._max_entity_length ~ 20-30, so it is fine to not vectorize this
            span_end_representations.append(
                torch.roll(token_representations, -shift, 1).unsqueeze(-2)
            )

            padding_mask = torch.ones(
                (batch_size, sequence_length), dtype=torch.bool, device=self.device
            )
            padding_mask[:, -shift:] = False
            padding_masks.append(padding_mask.unsqueeze(-1))

        padding_mask = torch.concat(padding_masks, dim=-1).bool()
        span_end_representation = torch.concat(span_end_representations, dim=-2)
        span_start_representation = token_representations.unsqueeze(-2).repeat(
            1, 1, self._max_entity_length, 1
        )

        span_length_embedding = self._get_length_representations(token_representations)

        span_representation = torch.concat(
            [span_start_representation, span_end_representation, span_length_embedding],
            dim=-1,
        )
        return self._span_projection(span_representation), padding_mask

    def _compute_sim_scores(
        self,
        representations: Tensor,
        entity_representations: Tensor,
        *,
        span_level: bool = True,
    ) -> Tensor:
        n_classes = len(self._encoded_descriptions)

        representations = representations.unsqueeze(
            -2
        )  # (B, S, M, 1, H) or (B, S, 1, H)
        new_shape = (
            (1, 1, 1, n_classes, self._hidden_size)
            if span_level
            else (1, 1, n_classes, self._hidden_size)
        )
        entity_representations = entity_representations.reshape(*new_shape)
        return self._metric(representations, entity_representations)

    def prepare_inputs(
        self,
        texts: List[str],
        entities: List[Optional[Set[TypedSpan]]],
        *,
        stride: float = 1 / 8,
        category_mapping: Dict[str, int],
        no_entity_category: str,
        text_lengths: Optional[List] = None,
    ) -> Iterable[Example]:
        encodings: Optional[List[EncodingFast]] = self._token_tokenizer(
            texts,
            truncation=False,
            add_special_tokens=False,
            return_offsets_mapping=True,
        ).encodings
        if encodings is None:
            raise ValueError(
                f"Tokenizer {self._token_tokenizer} is not fast! Use fast tokenizer!"
            )

        if text_lengths is not None:
            for i in range(len(texts)):
                text_lengths[i] = len(encodings[i].ids)

        for text_idx, (encoding, entities) in enumerate(zip(encodings, entities)):
            yield from strided_split(
                text_idx,
                encoding,
                entities,
                stride=int(self._max_sequence_length * stride),
                max_sequence_length=self._max_sequence_length,
                max_entity_length=self._max_entity_length,
                category_mapping=category_mapping,
                no_entity_category=no_entity_category,
                cls_token_id=self._token_tokenizer.cls_token_id,
            )

    def collate_examples(
        self, examples: Iterable[Example], return_batch_examples: bool = False
    ) -> Dict[str, Optional[LongTensor]]:
        return collate_examples(
            examples,
            padding_token_id=self._token_tokenizer.pad_token_id,
            pad_length=self._max_sequence_length,
            return_batch_examples=return_batch_examples,
        )

    def forward(
        self, input_ids: LongTensor, labels: Optional[LongTensor] = None, **_
    ) -> Union[LongTensor, Tuple[Tensor, LongTensor]]:
        """Predicts entity type ids. If true label ids are given,
        calculates loss as well."""

        batch_size, sequence_length = input_ids.shape
        input_ids = pad(
            input_ids,
            [0, self._max_sequence_length - sequence_length],
            value=self._token_tokenizer.pad_token_id,
        )

        token_representations = self._token_encoder(
            input_ids=input_ids.to(self.device)
        )[
            "last_hidden_state"
        ]  # (B, S, E)
        entity_representations = self._get_entity_representations()  # (C, E)

        span_representations, span_padding = self._get_span_representations(
            token_representations
        )
        span_scores = self._compute_sim_scores(
            span_representations, self._entity_projection(entity_representations)
        )
        batch_size, _, _, n_classes = span_scores.shape

        entity_threshold = span_scores[:, 0, 0].reshape(
            batch_size, 1, 1, n_classes
        )  # is a sim score with [CLS] token and entities
        mask = span_scores > entity_threshold
        scores_mask = torch.full_like(span_scores, fill_value=-torch.inf)
        scores_mask[mask] = 0  # set to -inf scores that did not pass the threshold

        values, predictions = torch.max(span_scores + scores_mask, dim=-1)
        predictions[values == -torch.inf] = self._unk_entity_label_id

        if labels is None:
            return predictions
        labels = labels.to(self.device)

        # (C, H)
        entity_start_representations = self._entity_start_projection(
            entity_representations
        )
        entity_end_representations = self._entity_end_projection(entity_representations)

        # (B, S, H)
        token_start_representations = self._token_start_projection(
            token_representations
        )
        token_end_representations = self._token_start_projection(token_representations)

        start_scores = self._compute_sim_scores(
            token_start_representations, entity_start_representations, span_level=False
        )
        end_scores = self._compute_sim_scores(
            token_end_representations, entity_end_representations, span_level=False
        )

        span_loss = self._loss_fn(span_scores, labels)
        start_loss = self._loss_fn(
            start_scores.unsqueeze(-2).repeat(1, 1, self._max_entity_length, 1), labels
        )
        end_loss = self._loss_fn(
            end_scores.unsqueeze(-2).repeat(1, 1, self._max_entity_length, 1), labels
        )

        return (
            self._span_coef * span_loss
            + self._start_coef * start_loss
            + self._end_coef * end_loss,
            predictions,
        )

    def save(self, save_path):
        super().save(save_path)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        cand_bert_path = os.path.join(save_path + "/cand_bert")
        cont_bert_path = os.path.join(save_path + "/cont_bert")
        if not os.path.exists(cand_bert_path):
            os.makedirs(cand_bert_path)
        if not os.path.exists(cont_bert_path):
            os.makedirs(cont_bert_path)
        self._context_encoder.save_pretrained(cont_bert_path)
        self._context_encoder.save_pretrained(cand_bert_path)
        self._descriptions_tokenizer.save_vocabulary(cand_bert_path)
        self._token_tokenizer.save_vocabulary(cont_bert_path)
        print(f"Saving non-shared model to : {save_path}")
        shutil.copyfile(
            os.path.join(
                Path(os.path.dirname(__file__)).parent, "config/t_config.json"
            ),
            os.path.join(save_path, "t_config.json"),
        )
        shutil.copyfile(
            os.path.join(
                Path(os.path.dirname(__file__)).parent, "config/m_config.json"
            ),
            os.path.join(save_path, "m_config.json"),
        )
        return f"[Saved model at] : {save_path}"

    def load(self, load_path) -> _Model:
        super().load(load_path)
        cand_bert_path = os.path.join(load_path, "cand_bert")
        cont_bert_path = os.path.join(load_path, "cont_bert")
        print(f"Loading non-shared model from : {load_path}")
        self._token_encoder: PreTrainedModel = AutoModel.from_pretrained(
            cont_bert_path, local_files_only=True
        )
        self._descriptions_encoder: PreTrainedModel = AutoModel.from_pretrained(
            cand_bert_path, local_files_only=True
        )
        self._token_tokenizer = AutoTokenizer.from_pretrained(
            cont_bert_path, local_files_only=True
        )
        self._descriptions_tokenizer = AutoTokenizer.from_pretrained(
            cand_bert_path, local_files_only=True
        )
        return self



import os
import json
from pathlib import Path
import shutil
from typing import List, Dict
import fasttext
import traceback
from fastapi import HTTPException
from jaseci.actions.live_actions import jaseci_action
from .json_to_train import json_to_train, prep_sentence, label_to_intent
from .config import (
    model_file_path,
    train_file_path,
    clf_json_file_path,
    base_json_file_path,
)

model = None
"""
Below code is to create a base_model.json file,
which help initialize a default model with a default data,
if not already exist.
"""
if not os.path.exists(base_json_file_path):
    default_data = {
        "sample_train_intent": ["this is a sample train text to create base model"]
    }
    json_object = json.dumps(default_data, indent=4)
    with open(base_json_file_path, "w") as outfile:
        outfile.write(json_object)


def updatetrainfile(traindata: Dict[str, str] = None, train_with_existing=True):
    data = {}
    """
    if we alreading have a existing training_data.json and we have to
    train_with_existing we append the data to the  training_data.json file
    otherwise we create training_data.json from the data provided
    """

    if os.path.exists(clf_json_file_path):
        with open(clf_json_file_path) as fp:
            data = json.load(fp)
    if train_with_existing is True:
        data.update(traindata)
    else:
        data = traindata
    json_object = json.dumps(data, indent=4)
    with open(clf_json_file_path, "w") as outfile:
        outfile.write(json_object)


@jaseci_action(act_group=["fast_enc"], allow_remote=True)
def train(traindata: Dict[str, List[str]] = None, train_with_existing: bool = True):
    global model
    print("Training...")
    # we pass the ##train_with_existing param to updatetrainfile function
    print(traindata)
    updatetrainfile(traindata, train_with_existing)

    json_to_train()
    model = fasttext.train_supervised(train_file_path, lr=0.25, epoch=30, wordNgrams=3)
    # print('Compressing...')
    # model.quantize(input=train_file_path, retrain=True)
    print("Saving...")
    model.save_model(model_file_path)
    print("")
    print(f"Model saved to {model_file_path}.")

    labels = [label.replace("__label__", "") for label in model.labels]
    print("")
    print(f"LABELS ({len(labels)}):")
    for label in labels:
        print(f"- {label}")
    if traindata is None:
        return model
    else:
        return "Model training Completed"


@jaseci_action(act_group=["fast_enc"], allow_remote=True)
def load_model(model_path: str = None):
    global model, model_file_path
    if model_path is not None:
        if type(model_path) is str:
            model_path = Path(model_path)
        model_file_path = os.path.join(model_path / "model.ftz")
    if os.path.exists(model_file_path):
        print("Model exists. Loading...")
        model = fasttext.load_model(model_file_path)
        print(f"Loaded {model_file_path}")
    else:
        print("Model does not exist. Training...")
        model = train()
    if model_path is None:
        return model
    else:
        return f"Model Loaded From : {model_path}"


@jaseci_action(act_group=["fast_enc"], allow_remote=True)
def save_model(model_path: str = None):
    if not model_path.isalnum():
        raise HTTPException(
            status_code=400,
            detail="Invalid model name. Only alphanumeric chars allowed.",
        )
    elif model is None:
        raise HTTPException(status_code=404, detail="Model has not been created ,yet!")
    print("Saving...")
    if type(model_path) is str:
        model_path = Path(model_path)
    model_path.mkdir(exist_ok=True, parents=True)
    state_save_path = os.path.join(model_path / "model.ftz")
    # model_path = (model_path / "model.ftz")
    model.save_model(state_save_path)
    # we also ship the training_data.json file
    # with the model to maintain the state of the model
    shutil.copyfile(base_json_file_path, model_path / "training_data.json")
    return f"Model saved to {model_path}."


@jaseci_action(act_group=["fast_enc"], allow_remote=True)
def predict(sentences: List[str]):
    global model
    try:
        model = model if model is not None else load_model()
        result = {}
        for sentence in sentences:
            result[sentence] = []
            predictions = model.predict(prep_sentence(sentence))
            for pre in zip(predictions[0], predictions[1]):
                result[sentence].append(
                    {
                        "sentence": sentence,
                        "intent": label_to_intent(pre[0]),
                        "probability": pre[1],
                    }
                )
        return result
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    print("FasttextClassifier up and running")
    launch_server(port=8000)



import os
import torch
from typing import Dict, List, Union
from fastapi import HTTPException
from transformers import AutoModel, AutoConfig, AutoTokenizer
import traceback
import numpy as np
from jaseci.actions.live_actions import jaseci_action
import random
import json
import shutil

from .utils.evaluate import get_embeddings  # noqa
from .utils.models import BiEncoder  # noqa
from .utils.train import train_model  # noqa


# device = torch.device("cpu")
# uncomment this if you wish to use GPU to train
# this is commented out because this causes issues with
# unittest on machines with GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


# funtion to set seed for the module
def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def setup():
    """
    Loading configurations from utils/config.cfg and
    initialize tokenizer and model
    """
    global model, tokenizer, model_config, train_config, t_config_fname, m_config_fname
    dirname = os.path.dirname(__file__)
    m_config_fname = os.path.join(dirname, "utils/model_config.json")
    t_config_fname = os.path.join(dirname, "utils/train_config.json")
    with open(m_config_fname, "r") as jsonfile:
        model_config = json.load(jsonfile)
    with open(t_config_fname, "r") as jsonfile:
        train_config = json.load(jsonfile)

    train_config.update({"device": device.type})
    trf_config = AutoConfig.from_pretrained(model_config["model_name"])
    tokenizer = AutoTokenizer.from_pretrained(
        model_config["model_name"], do_lower_case=True, clean_text=False
    )
    if model_config["shared"] is True:
        cont_bert = AutoModel.from_config(trf_config)
        cand_bert = cont_bert
        print("shared model created")
    else:
        cont_bert = AutoModel.from_config(trf_config)
        cand_bert = AutoModel.from_config(trf_config)
        print("non shared model created")
    model = BiEncoder(
        config=trf_config,
        cont_bert=cont_bert,
        cand_bert=cand_bert,
        shared=model_config["shared"],
        loss_type=model_config["loss_type"],
        loss_function=model_config["loss_function"],
    )

    model.to(train_config["device"])
    set_seed(train_config["seed"])


setup()


# API for getting the cosine similarity
@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def cosine_sim(vec_a: List[float], vec_b: List[float]):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """

    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return result.astype(float)


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def dot_prod(vec_a: List[float], vec_b: List[float]):
    """
    Caculate the dot product of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - dot product
    """
    dot_product = np.matmul(vec_a, vec_b)
    return dot_product.astype(float)


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def infer(
    contexts: Union[List[str], List[List]],
    candidates: Union[List[str], List[List]],
    context_type: str,
    candidate_type: str,
):
    """
    Take list of context, candidate and return nearest candidate to the context
    """
    model.eval()
    predicted_candidates = []
    try:
        if (context_type == "text") and (candidate_type == "text"):
            con_embed = []
            con_embed = get_context_emb(contexts)
            cand_embed = get_candidate_emb(candidates)
        elif (context_type == "text") and (candidate_type == "embedding"):
            con_embed = get_context_emb(contexts)
            cand_embed = candidates
        elif (context_type == "embedding") and (candidate_type == "text"):
            con_embed = contexts
            cand_embed = get_candidate_emb(candidates)

        elif (context_type == "embedding") and (candidate_type == "embedding"):
            con_embed = contexts
            cand_embed = candidates
        else:
            raise HTTPException(status_code=404, detail=str("input type not supported"))
        for data, cont in zip(con_embed, contexts):
            score_dat = []
            out_data = {"context": str, "candidate": [], "score": [], "predicted": {}}
            if candidate_type == "embedding":
                for lbl in cand_embed:
                    if model_config["loss_type"] == "cos":
                        score_dat.append(cosine_sim(vec_a=data, vec_b=lbl))
                out_data["predicted"] = {
                    "label": candidates[
                        out_data["score"].index(max(out_data["score"]))
                    ],
                    "score": max(out_data["score"]),
                }
                predicted_candidates.append(int(np.argmax(score_dat)))
            else:
                for lbl, cand in zip(cand_embed, candidates):
                    if model_config["loss_type"] == "cos":
                        out_data["context"] = cont
                        out_data["candidate"].append(cand)
                        out_data["score"].append(cosine_sim(vec_a=data, vec_b=lbl))
                    else:
                        out_data["context"] = cont
                        out_data["candidate"].append(cand)
                        out_data["score"].append(float(dot_prod(vec_a=data, vec_b=lbl)))
                out_data["predicted"] = {
                    "label": candidates[
                        out_data["score"].index(max(out_data["score"]))
                    ],
                    "score": max(out_data["score"]),
                }
                predicted_candidates.append(out_data)
        return predicted_candidates
    except Exception as e:
        raise HTTPException(
            status_code=404,
            detail=str(
                f"""input type can be
         'embedding' or 'text', context and
        candidate type should match the content of contexts and candidates.
        Exception : {e}"""
            ),
        )


# API for training
@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def train(dataset: Dict = None, from_scratch=False, training_parameters: Dict = None):
    """
    Take list of context, candidate, labels and trains the model
    """
    global model
    train_data = {"contexts": [], "candidates": [], "labels": []}
    if from_scratch is True:
        save_model(model_config["model_save_path"])
        setup()
    model.train()
    try:
        if training_parameters is not None:
            with open(t_config_fname, "w+") as jsonfile:
                train_config.update(training_parameters)
                json.dump(train_config, jsonfile, indent=4)
        for data in dataset.keys():
            for dat in dataset[data]:
                train_data["contexts"].append(dat)
                train_data["candidates"].append(data)
                train_data["labels"].append(1)
        model = train_model(
            model=model,
            tokenizer=tokenizer,
            contexts=train_data["contexts"],
            candidates=train_data["candidates"],
            labels=train_data["labels"],
            train_config=train_config,
        )
        return "Model Training is complete."
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


# API for geting Context Embedding
@jaseci_action(act_group=["bi_enc"], aliases=["encode_context"], allow_remote=True)
def get_context_emb(contexts: List):
    """
    Take list of context and returns the embeddings
    """
    model.eval()
    embedding = []
    for cont in contexts:
        embedding.append(
            get_embeddings(
                model=model,
                tokenizer=tokenizer,
                text_data=cont,
                embed_type="context",
                train_config=train_config,
            )
        )
    return embedding


# API for geting Candidates Embedding


@jaseci_action(act_group=["bi_enc"], aliases=["encode_candidate"], allow_remote=True)
def get_candidate_emb(candidates: List):
    """
    Take list of candidates and returns the embeddings
    """
    model.eval()
    embedding = get_embeddings(
        model,
        tokenizer,
        text_data=candidates,
        embed_type="candidate",
        train_config=train_config,
    )
    return embedding


# API for setting the training and model parameters
@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def get_train_config():
    try:
        with open(t_config_fname, "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def set_train_config(training_parameters: Dict = None):
    global train_config
    try:
        with open(t_config_fname, "w+") as jsonfile:
            train_config.update(training_parameters)
            json.dump(train_config, jsonfile, indent=4)
        return "Config setup is complete."
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def get_model_config():
    try:
        with open(m_config_fname, "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def set_model_config(model_parameters: Dict = None):
    global model_config
    try:
        save_model(model_config["model_save_path"])
        with open(m_config_fname, "w+") as jsonfile:
            model_config.update(model_parameters)
            json.dump(model_config, jsonfile, indent=4)

        setup()
        return "Config setup is complete."

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def save_model(model_path: str):
    """
    saves the model to the provided model_path
    """
    try:
        if not model_path.replace("_", "").isalnum():
            raise HTTPException(
                status_code=400,
                detail="""
                Invalid model name. Model Name can only have Alphanumeric
                 and '_' characters.""",
            )
        if not os.path.exists(model_path):
            os.makedirs(model_path)
        if model_config["shared"] is True:
            model.cont_bert.save_pretrained(model_path)
            tokenizer.save_vocabulary(model_path)
            print(f"Saving shared model to : {model_path}")
        else:
            cand_bert_path = os.path.join(model_path + "/cand_bert")
            cont_bert_path = os.path.join(model_path + "/cont_bert")
            if not os.path.exists(cand_bert_path):
                os.makedirs(cand_bert_path)
            if not os.path.exists(cont_bert_path):
                os.makedirs(cont_bert_path)
            tokenizer.save_vocabulary(cand_bert_path)
            tokenizer.save_vocabulary(cont_bert_path)
            model.cont_bert.save_pretrained(cont_bert_path)
            model.cand_bert.save_pretrained(cand_bert_path)
            print(f"Saving non-shared model to : {model_path}")
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "utils/train_config.json"),
            os.path.join(model_path, "train_config.json"),
        )
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "utils/model_config.json"),
            os.path.join(model_path, "model_config.json"),
        )
        return f"[Saved model at] : {model_path}"
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def load_model(model_path):
    """
    loads the model from the provided model_path
    """
    global model, tokenizer
    if not os.path.exists(model_path):
        raise HTTPException(status_code=404, detail="Model path is not available")
    try:
        with open(m_config_fname, "r") as jsonfile:
            model_config_data = json.load(jsonfile)
        if model_config_data["shared"] is True:
            trf_config = AutoConfig.from_pretrained(model_path, local_files_only=True)
            tokenizer = AutoTokenizer.from_pretrained(
                model_path, do_lower_case=True, clean_text=False
            )
            cont_bert = AutoModel.from_pretrained(model_path, local_files_only=True)
            cand_bert = cont_bert
            print(f"Loading shared model from : {model_path}")
        else:
            cand_bert_path = os.path.join(model_path, "cand_bert")
            cont_bert_path = os.path.join(model_path, "cont_bert")
            print(f"Loading non-shared model from : {model_path}")
            cont_bert = AutoModel.from_pretrained(cont_bert_path, local_files_only=True)
            cand_bert = AutoModel.from_pretrained(cand_bert_path, local_files_only=True)
            trf_config = AutoConfig.from_pretrained(
                cont_bert_path, local_files_only=True
            )
            tokenizer = AutoTokenizer.from_pretrained(
                cand_bert_path, do_lower_case=True, clean_text=False
            )
        model = BiEncoder(
            config=trf_config,
            cont_bert=cont_bert,
            cand_bert=cand_bert,
            shared=model_config_data["shared"],
            loss_type=model_config["loss_type"],
            loss_function=model_config["loss_function"],
        )
        model.to(train_config["device"])
        return f"[loaded model from] : {model_path}"
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from abc import abstractmethod, ABCMeta

from torch import Tensor
from torch.nn import CosineSimilarity, Module


class Metric(Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        pass


class CosSimilarity(Module):
    def __init__(self, scale: float):
        super().__init__()
        self._scale = scale
        self._metric = CosineSimilarity(dim=-1)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        return self._metric(x, y) / self._scale



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module


class ClSummerModule(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(ClSummerModule, cls).setUpClass()
        ret = load_module_actions("jac_nlp.cl_summer")
        assert ret == True

    @jac_testcase("cl_summer.jac", "test_summarizer_text")
    def test_summarizer_text(self, ret):
        self.assertEqual(len(ret["report"][0]), 1)

    @jac_testcase("cl_summer.jac", "test_summarizer_url")
    def test_summarizer_url(self, ret):
        self.assertEqual(len(ret["report"][0]), 5)

    @classmethod
    def tearDownClass(cls):
        super(ClSummerModule, cls).tearDownClass()
        ret = unload_module("jac_nlp.cl_summer.cl_summer")
        assert ret == True



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module
import pytest


class T5SumTest(CoreTest):
    fixture_src = __file__
    """
    Test Class for T5Sum Module to test the functionality of api's
    """

    @classmethod
    def setUpClass(cls):
        super(T5SumTest, cls).setUpClass()
        ret = load_module_actions("jac_nlp.t5_sum")
        assert ret is True

    @pytest.mark.order(1)
    @jac_testcase("t5_sum.jac", "test_t5_sum_detection")
    def test_t5_sum_detection(self, ret):
        self.assertEqual(ret["success"], True)

    @classmethod
    def tearDownClass(cls):
        super(T5SumTest, cls).tearDownClass()
        ret = unload_module("jac_nlp.t5_sum.t5_sum")
        assert ret is True



from transformers import (
    pipeline,
    GPT2Tokenizer,
    GPT2Model,
    AutoTokenizer,
    AutoModelWithLMHead,
)
import torch
from jaseci.actions.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
from typing import List, Union
import os
from .train import prepare_data, load_dataset, get_trainer
import shutil

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

model = None
tokenizer = None
generator = None


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def setup(model_name: str = "gpt2", get_embeddings: bool = False):
    global model, tokenizer, generator
    if get_embeddings:
        generator = None
        tokenizer = GPT2Tokenizer.from_pretrained(model_name)
        tokenizer.pad_token = tokenizer.eos_token
        model = GPT2Model.from_pretrained(model_name).to(device)
    else:
        model, tokenizer = None, None
        generator = pipeline("text-generation", model=model_name, tokenizer="gpt2")


setup(model_name="gpt2", get_embeddings=False)


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def generate(
    text: Union[List[str], str],
    max_length: int = 30,
    min_length: int = 10,
    num_return_sequences: int = 3,
) -> List:
    global generator, model, tokenizer
    if generator is None:
        model, tokenizer = None, None
        model_name = "./gpt2-trained" if os.path.exists("gpt2-trained") else "gpt2"
        setup(model_name=model_name, get_embeddings=False)
    try:
        if isinstance(text, str):
            text = [text]
        return generator(
            text,
            max_length=max_length,
            min_length=min_length,
            num_return_sequences=num_return_sequences,
        )
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def get_embeddings(text: Union[List[str], str]) -> List:
    global model, tokenizer, generator, device
    if model is None or tokenizer is None:
        generator = None
        setup(model_name="gpt2", get_embeddings=True)
    try:
        if isinstance(text, str):
            text = [text]
        inputs = tokenizer(text, return_tensors="pt", padding=True).to(device)
        with torch.no_grad():
            outputs = model(**inputs)
        return outputs.last_hidden_state.tolist()
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def train(
    text: Union[List[str], str], epochs: int = 1, use_prev_trained=True, freeze=True
):
    if not use_prev_trained:
        shutil.rmtree("gpt2-trained", ignore_errors=True)
    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    train_path, test_path = prepare_data(text)
    train_dataset, test_dataset, data_collator = load_dataset(
        train_path, test_path, tokenizer
    )
    model_name = (
        "./gpt2-trained"
        if os.path.exists("gpt2-trained") and use_prev_trained
        else "gpt2"
    )
    model = AutoModelWithLMHead.from_pretrained(model_name)

    if freeze:
        for param in model.parameters():
            param.requires_grad = False
        model.lm_head.weight.requires_grad = True

    trainer = get_trainer(model, train_dataset, test_dataset, data_collator, epochs)
    trainer.train()
    trainer.save_model()

    global generator
    generator = pipeline("text-generation", model="./gpt2-trained", tokenizer="gpt2")


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module
import pytest


class BiEncTest(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(BiEncTest, cls).setUpClass()
        ret = load_module_actions("jac_nlp.bi_enc")
        assert ret is True

    @pytest.mark.order(1)
    @jac_testcase("bi_enc.jac", "test_bi_enc_cos_sim")
    def test_cos_sim_function(self, ret):
        self.assertEqual(round(ret["report"][0], 2), 0.98)

    @pytest.mark.order(2)
    @jac_testcase("bi_enc.jac", "test_bi_enc_infer")
    def test_biencoder_infer(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(3)
    @jac_testcase("bi_enc.jac", "test_bi_enc_context_emb")
    def test_biencoder_context_emb(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(len(ret["report"][0][0]), 128)

    @pytest.mark.order(4)
    @jac_testcase("bi_enc.jac", "test_bi_enc_cand_emb")
    def test_biencoder_candidate_emb(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(len(ret["report"][0][0]), 128)

    @pytest.mark.order(5)
    @jac_testcase("bi_enc.jac", "test_bi_enc_get_train_config")
    def test_biencoder_get_train_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(6)
    @jac_testcase("bi_enc.jac", "test_bi_enc_get_model_config")
    def test_biencoder_get_model_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(7)
    @jac_testcase("bi_enc.jac", "test_bi_enc_train")
    def test_biencoder_train(self, ret):
        self.assertEqual(ret["report"][0], "Model Training is complete.")
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(8)
    @jac_testcase("bi_enc.jac", "test_bi_enc_save_model")
    def test_biencoder_save_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(9)
    @jac_testcase("bi_enc.jac", "test_bi_enc_load_model")
    def test_biencoder_load_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(10)
    @jac_testcase("bi_enc.jac", "test_bi_enc_set_model_config")
    def test_biencoder_set_model_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(11)
    @jac_testcase("bi_enc.jac", "test_bi_enc_set_train_config")
    def test_biencoder_set_train_config(self, ret):
        self.assertEqual(ret["success"], True)

    @classmethod
    def tearDownClass(cls):
        super(BiEncTest, cls).tearDownClass()
        ret = unload_module("jac_nlp.bi_enc.bi_enc")
        assert ret is True



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module
import pytest


class EntExtTest(CoreTest):
    fixture_src = __file__
    """
    Test Class for EntExt Module to test the functionality of api's
    """

    @classmethod
    def setUpClass(cls):
        super(EntExtTest, cls).setUpClass()
        ret = load_module_actions("jac_nlp.ent_ext")
        assert ret is True

    @pytest.mark.order(1)
    @jac_testcase("ent_ext.jac", "test_ent_ext_set_config")
    def test_ent_ext_set_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(2)
    @jac_testcase("ent_ext.jac", "test_train_ner")
    def test_train_ner(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(3)
    @jac_testcase("ent_ext.jac", "test_extract_entity")
    def test_extract_entity(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(4)
    @jac_testcase("ent_ext.jac", "test_ent_ext_save_model")
    def test_ent_ext_save_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("ent_ext.jac", "test_ent_ext_load_model")
    def test_ent_ext_load_model(self, ret):
        self.assertEqual(ret["success"], True)

    @classmethod
    def tearDownClass(cls):
        super(EntExtTest, cls).tearDownClass()
        ret = unload_module("jac_nlp.ent_ext.ent_ext")
        assert ret is True



import tempfile
from sklearn.model_selection import train_test_split
from transformers import (
    TextDataset,
    DataCollatorForLanguageModeling,
    Trainer,
    TrainingArguments,
)
from typing import List, Union


def prepare_data(text: Union[List[str], str], split: float = 0.2):
    if isinstance(text, str):
        text = text.strip().split()
        train, test = train_test_split(text, test_size=split, shuffle=False)
        train, test = " ".join(train), " ".join(test)
    elif isinstance(text, list):
        train, test = train_test_split(text, test_size=split)
        train, test = " ".join(train), " ".join(test)
    else:
        raise Exception("Text must be a string or list of strings")

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(train)
        train_path = f.name
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(test)
        test_path = f.name
    return train_path, test_path


def load_dataset(train_path, test_path, tokenizer):
    train_dataset = TextDataset(
        tokenizer=tokenizer, file_path=train_path, block_size=128
    )

    test_dataset = TextDataset(tokenizer=tokenizer, file_path=test_path, block_size=128)

    data_collator = DataCollatorForLanguageModeling(
        tokenizer=tokenizer,
        mlm=False,
    )
    return train_dataset, test_dataset, data_collator


def get_trainer(model, train_dataset, test_dataset, data_collator, epochs=1):
    training_args = TrainingArguments(
        output_dir="./gpt2-trained",  # The output directory
        overwrite_output_dir=True,  # overwrite the content of the output directory
        num_train_epochs=epochs,  # number of training epochs
        per_device_train_batch_size=32,  # batch size for training
        per_device_eval_batch_size=64,  # batch size for evaluation
        eval_steps=400,  # Number of update steps between two evaluations.
        save_steps=800,  # after # steps model is saved
        warmup_steps=500,  # number of warmup steps for learning rate scheduler
        prediction_loss_only=True,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=data_collator,
        train_dataset=train_dataset,
        eval_dataset=test_dataset,
    )

    return trainer



import torch
from torch.utils.data import Dataset


# this class transforms data in required training format
# and inference for text and tokens
class SelectionDataset(Dataset):
    def __init__(
        self,
        contexts,
        candidates,
        context_transform,
        candidate_transform,
        labels=None,
        mode="train",
    ):
        self.context_transform = context_transform
        self.candidate_transform = candidate_transform
        self.data_source = []
        self.mode = mode
        if mode == "eval":
            for text in contexts:
                group = {"context": None, "candidates": [], "labels": []}
                for cand in candidates:
                    group["candidates"].append(cand)
                    # below label is 0 for all,used for help in inference
                    group["labels"].append(0)
                    group["context"] = [text]
                self.data_source.append(group)
        else:
            group = {"context": None, "candidates": [], "labels": []}
            for text, cand, lbl in zip(contexts, candidates, labels):
                group = {"context": None, "candidates": [], "labels": []}
                # below code is to combine positive and negetive sample
                # if lbl == 1 and len(group['candidates']) > 0:
                #     self.data_source.append(group)
                #     group = {
                #         'context': None,
                #         'candidates': [],
                #         'labels': []
                #     }
                group["candidates"].append(cand)
                group["labels"].append(lbl)
                group["context"] = [text]
                # if len(group['candidates']) > 0:
                self.data_source.append(group)
        group = {"context": None, "candidates": [], "labels": []}
        if len(self.data_source) < 2 and mode != "eval":
            group["context"] = ["This is sample text"]
            group["candidates"].append("sampletext")
            group["labels"].append(1)
            if len(candidates) > 1:
                group["context"] = ["This is a sample text"]
                group["candidates"].append("notsampletext")
                group["labels"].append(0)
            self.data_source.append(group)

    def __len__(self):
        return len(self.data_source)

    def __getitem__(self, index):
        group = self.data_source[index]
        context, candidates, labels = (
            group["context"],
            group["candidates"],
            group["labels"],
        )

        transformed_context = self.context_transform(
            context
        )  # [token_ids],[seg_ids],[masks]
        transformed_candidates = self.candidate_transform(
            candidates
        )  # [token_ids],[seg_ids],[masks]
        ret = transformed_context, transformed_candidates, labels

        return ret

    def batchify_join_str(self, batch):
        (
            contexts_token_ids_list_batch,
            contexts_input_masks_list_batch,
            candidates_token_ids_list_batch,
            candidates_input_masks_list_batch,
        ) = ([], [], [], [])
        labels_batch = []
        for sample in batch:
            (contexts_token_ids_list, contexts_input_masks_list), (
                candidates_token_ids_list,
                candidates_input_masks_list,
            ) = sample[:2]

            contexts_token_ids_list_batch.append(contexts_token_ids_list)
            contexts_input_masks_list_batch.append(contexts_input_masks_list)

            candidates_token_ids_list_batch.append(candidates_token_ids_list)
            candidates_input_masks_list_batch.append(candidates_input_masks_list)

            labels_batch.append(sample[-1])

        long_tensors = [
            contexts_token_ids_list_batch,
            contexts_input_masks_list_batch,
            candidates_token_ids_list_batch,
            candidates_input_masks_list_batch,
        ]

        (
            contexts_token_ids_list_batch,
            contexts_input_masks_list_batch,
            candidates_token_ids_list_batch,
            candidates_input_masks_list_batch,
        ) = (torch.tensor(t, dtype=torch.long) for t in long_tensors)

        labels_batch = torch.tensor(labels_batch, dtype=torch.float)
        return (
            contexts_token_ids_list_batch,
            contexts_input_masks_list_batch,
            candidates_token_ids_list_batch,
            candidates_input_masks_list_batch,
            labels_batch,
        )


# this class transforms data to generate embeddings
class EvalDataset(Dataset):
    def __init__(
        self, texts, context_transform=None, candidate_transform=None, mode="context"
    ):
        self.context_transform = context_transform
        self.candidate_transform = candidate_transform
        self.data_source = []
        self.mode = mode
        group = {"text": []}
        if mode == "context":
            group["text"].append(texts)
        else:
            for text in texts:
                group["text"].append(text)
        self.data_source.append(group)

    def __len__(self):
        return len(self.data_source)

    def __getitem__(self, index):
        group = self.data_source[index]
        text = group["text"]
        if self.mode == "context":
            transformed_text = self.context_transform(text)  # [token_ids],[masks]
        else:
            transformed_text = self.candidate_transform(text)  # [token_ids],[masks]

        return transformed_text

    def eval_str(self, batch):
        token_ids_list_batch, input_masks_list_batch = [], []
        for sample in batch:
            token_ids_list, input_masks_list = sample
            token_ids_list_batch.append(token_ids_list)
            input_masks_list_batch.append(input_masks_list)

        long_tensors = [token_ids_list_batch, input_masks_list_batch]
        token_ids_list_batch, input_masks_list_batch = (
            torch.tensor(t, dtype=torch.long) for t in long_tensors
        )
        return token_ids_list_batch, input_masks_list_batch


# this class is for creating token data for candidate
class SelectionSequentialTransform(object):
    def __init__(self, tokenizer, max_len):
        self.tokenizer = tokenizer
        self.max_len = max_len

    def __call__(self, texts):
        input_ids_list, input_masks_list = [], []
        for text in texts:
            tokenized_dict = self.tokenizer.encode_plus(
                text, padding="max_length", max_length=self.max_len, truncation=True
            )
            input_ids, input_masks = (
                tokenized_dict["input_ids"],
                tokenized_dict["attention_mask"],
            )
            assert len(input_ids) == self.max_len
            assert len(input_masks) == self.max_len
            input_ids_list.append(input_ids)
            input_masks_list.append(input_masks)
        return input_ids_list, input_masks_list


# this class is for creating token data for context
class SelectionJoinTransform(object):
    def __init__(self, tokenizer, max_len):
        self.tokenizer = tokenizer
        self.max_len = max_len

        self.cls_id = self.tokenizer.convert_tokens_to_ids("[CLS]")
        self.sep_id = self.tokenizer.convert_tokens_to_ids("[SEP]")
        self.pad_id = 0

    def __call__(self, texts):
        # another option is to use [SEP], but here we follow the discussion at:
        # https://github.com/facebookresearch/ParlAI/issues/2306#issuecomment-599180186
        context = "\n".join(texts)
        tokenized_dict = self.tokenizer.encode_plus(context)
        input_ids, input_masks = (
            tokenized_dict["input_ids"],
            tokenized_dict["attention_mask"],
        )
        input_ids = input_ids[-self.max_len :]
        input_ids[0] = self.cls_id
        input_masks = input_masks[-self.max_len :]
        input_ids += [self.pad_id] * (self.max_len - len(input_ids))
        input_masks += [0] * (self.max_len - len(input_masks))
        assert len(input_ids) == self.max_len
        assert len(input_masks) == self.max_len

        return input_ids, input_masks



import traceback
from fastapi import HTTPException
from jaseci.actions.live_actions import jaseci_action
from typing import Dict, List, Optional
import os
import json
import warnings
from .train import (
    predict_text,
    load_custom_model,
    save_custom_model,
    NERDataMaker,
    train_model,
)
import shutil

warnings.filterwarnings("ignore")


def config_setup():
    """
    Loading configurations from config and
    initialize tokenizer and model
    """
    global train_config, model_config, curr_model_path
    global t_config_fname, m_config_fname
    dirname = os.path.dirname(__file__)
    m_config_fname = os.path.join(dirname, "utils/model_config.json")
    t_config_fname = os.path.join(dirname, "utils/train_config.json")

    with open(t_config_fname, "r") as jsonfile:
        train_config = json.load(jsonfile)
    with open(m_config_fname, "r") as jsonfile:
        model_config = json.load(jsonfile)

    curr_model_path = model_config["model_name"]


config_setup()

enum = {"default": 1, "append": 2, "incremental": 3}


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def extract_entity(text: str = None):
    try:
        data = predict_text(text)
        if isinstance(data, list):
            return data
        else:
            return "No active model found. Please train or load model first before inference."
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def train(
    mode: str = train_config["MODE"],
    epochs: int = train_config["EPOCHS"],
    train_data: List = [],
    val_data: Optional[List] = [],
):
    """
    API for training the model
    """
    if not os.path.exists("train/logs"):
        os.makedirs("train/logs")
    if mode == "default":
        train_config["EPOCHS"] = epochs
        with open("train/train.json", "w") as fp:
            json.dump(train_data, fp)
    elif mode == "incremental":
        train_config["EPOCHS"] = train_config["EPOCHS"] + epochs
        with open("train/train.json", "w") as fp:
            json.dump(train_data, fp)
    elif mode == "append":
        train_config["EPOCHS"] = epochs
        if os.path.exists("train/train.json"):
            with open("train/train.json", "r") as fp:
                old_train_data = json.load(fp)
            for data in old_train_data:
                train_data.append(data)
            with open("train/train.json", "w") as fp:
                json.dump(train_data, fp)

    else:
        st = {
            "Status": "training failed",
            "Error": "Please define training mode",
            "Available mode": ["default", "append", "incremental"],
        }
        raise HTTPException(status_code=400, detail=st)

    try:
        train_dm = NERDataMaker(train_data)
        load_custom_model(curr_model_path, train_dm)
        train_model(train_data=train_data, val_data=val_data, train_config=train_config)
        print("model training Completed")
        return {"status": "model Training Successful!"}
    except Exception as e:
        print(e)
        traceback.print_exc()
        raise HTTPException(
            status_code=500,
            detail=str("Issue encountered during train "),
        )


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def load_model(model_path: str = "default", local_file: bool = True):
    global curr_model_path
    curr_model_path = model_path
    train_file_path = os.path.join(model_path, "train.json")
    if local_file is True and not os.path.exists(model_path):
        return "Model path is not available"
    try:
        print("loading latest trained model to memory...")
        if os.path.exists(train_file_path):
            with open(train_file_path, "r") as fp:
                train_data = json.load(fp)
            train_dm = NERDataMaker(train_data)
            load_custom_model(curr_model_path, train_dm)
            print("model successfully load to memory!")
            return {"status": "model Loaded Successfull!"}
        else:
            raise HTTPException(status_code=400, detail=str("cannot load model"))
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=400, detail=str(e))


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def save_model(model_path: str = "mymodel"):
    try:
        save_custom_model(model_path)
        print(f"current model {model_path} saved to disk.")
        shutil.copy("train/train.json", model_path)
        return {"status": f"model {model_path} saved Successfull!"}
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=400, detail=str(e))


# API for setting the training and model parameters
@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def get_train_config():
    try:
        with open(t_config_fname, "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def set_train_config(training_parameters: Dict = None):
    global train_config
    try:
        train_config.update(training_parameters)
        with open(t_config_fname, "w+") as jsonfile:
            json.dump(train_config, jsonfile, indent=4)

        return "Config setup is complete."
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def get_model_config():
    try:
        with open(m_config_fname, "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["tfm_ner"], allow_remote=True)
def set_model_config(model_parameters: Dict = None):
    global model_config
    try:
        save_model(model_config["model_save_path"])
        model_config.update(model_parameters)
        with open(m_config_fname, "w+") as jsonfile:
            json.dump(model_config, jsonfile, indent=4)

        config_setup()
        return "Config setup is complete."
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from enum import Enum
from typing import TypeVar, Dict, List, Any, Optional, Tuple

# from sklearn.metrics import f1_score, recall_score, precision_score
# from transformers import EvalPrediction
import numpy as np
import torch
from torch import Tensor
from torch.nn.functional import pad

_K = TypeVar("_K")
_V = TypeVar("_V")


def invert(d: Dict[_K, _V]) -> Dict[_V, _K]:
    return {v: k for k, v in d.items()}


_Tensor = TypeVar("_Tensor", bound=Tensor)


def pad_images(
    images: List[_Tensor],
    *,
    padding_value: Any = 0.0,
    padding_length: Tuple[Optional[int], Optional[int]],
) -> _Tensor:
    """Pad images to equal length (maximum height and width)."""
    max_height, max_width = padding_length

    shapes = torch.tensor(
        list(map(lambda t: t.shape, images)), dtype=torch.long
    ).transpose(0, 1)
    max_height = shapes[-2].max() if max_height is None else max_height
    max_width = shapes[-1].max() if max_width is None else max_width

    ignore_dims = len(images[0].shape) - 2

    image_batch = [
        # The needed padding is the difference between the
        # max width/height and the image's actual width/height.
        pad(
            img,
            [
                *([0, 0] * ignore_dims),
                0,
                max_width - img.shape[-1],
                0,
                max_height - img.shape[-2],
            ],
            value=padding_value,
        )
        for img in images
    ]
    return torch.stack(image_batch)


def to_numpy(tensor: Tensor) -> np.ndarray:
    return tensor.detach().cpu().numpy()


class DatasetType(str, Enum):
    TRAIN = "train"
    DEV = "dev"
    TEST = "test"


# def compute_metrics(
#     evaluation_results: EvalPrediction,
#     category_id_mapping: Dict[int, str],
#     no_entity_category_id: int,
# ) -> Dict[str, float]:

#     padding_mask = evaluation_results.label_ids != -100

#     label_ids = evaluation_results.label_ids[padding_mask]
#     predictions = evaluation_results.predictions[padding_mask]

#     unique_label_ids = set(np.unique(label_ids[label_ids != no_entity_category_id]))

#     labels = sorted(category_id_mapping.keys())
#     f1_category_scores = f1_score(
#         label_ids, predictions, average=None, labels=labels, zero_division=0
#     )
#     recall_category_scores = recall_score(
#         label_ids, predictions, average=None, labels=labels, zero_division=0
#     )
#     precision_category_scores = precision_score(
#         label_ids, predictions, average=None, labels=labels, zero_division=0
#     )
#     results: Dict[str, float] = {}
#     sum_f1 = 0
#     sum_recall = 0
#     sum_precision = 0
#     for category_id, f1, recall, precision in zip(
#         labels, f1_category_scores, recall_category_scores, precision_category_scores
#     ):
#         if category_id == no_entity_category_id:
#             # logger.info(f'O: {f1}, {recall}, {precision}')
#             continue

#         if category_id not in unique_label_ids:
#             # logger.info(f'Skipping {category_id}: {f1}, {recall}, {precision}')
#             continue

#         category = category_id_mapping[category_id]
#         results[f"F1_{category}"] = f1
#         results[f"Recall_{category}"] = recall
#         results[f"Precision_{category}"] = precision
#         sum_f1 += f1
#         sum_recall += recall
#         sum_precision += precision

#     num_categories = len(category_id_mapping) - 1

#     results["F1_macro"] = sum_f1 / num_categories
#     results["Recall_macro"] = sum_recall / num_categories
#     results["Precision_macro"] = sum_precision / num_categories
#     return results


def get_category_id_mapping(model_args: Dict, category_name=None):
    category_names = sorted(category_name)
    category_id_mapping = dict(enumerate(category_names))
    category_id_mapping[model_args["unk_entity_type_id"]] = model_args["unk_category"]
    return category_id_mapping



import numpy as np
import tensorflow_hub as hub
import tensorflow as tf
import tensorflow_text  # noqa
from jaseci.actions.live_actions import jaseci_action
from typing import Union


module = hub.load("https://tfhub.dev/google/universal-sentence-encoder/4")


@jaseci_action(act_group=["use"], aliases=["get_embedding"], allow_remote=True)
def encode(text: Union[str, list]):
    if isinstance(text, str):
        text = [text]
    return module(text).numpy().tolist()


@jaseci_action(act_group=["use"], allow_remote=True)
def cos_sim_score(q_emb: list, a_emb: list):
    norm = np.linalg.norm
    result = np.dot(q_emb, a_emb) / (norm(q_emb) * norm(a_emb))
    return float(result.astype(float))


@jaseci_action(act_group=["use"], allow_remote=True)
def text_similarity(text1: str, text2: str):
    enc_a = np.squeeze(np.asarray(encode(text1)))
    enc_b = np.squeeze(np.asarray(encode(text2)))
    return cos_sim_score(list(enc_a), list(enc_b))


@jaseci_action(act_group=["use"], allow_remote=True)
def text_classify(text: str, classes: list):
    text_emb = np.squeeze(np.asarray(encode(text)))
    ret = {"match": "", "match_idx": -1, "scores": []}
    for i in classes:
        i_emb = np.squeeze(np.asarray(encode(i))) if isinstance(i, str) else i
        ret["scores"].append(cos_sim_score(text_emb, i_emb))
    top_hit = ret["scores"].index(max(ret["scores"]))
    ret["match_idx"] = top_hit
    ret["match"] = (
        classes[top_hit] if isinstance(classes[top_hit], str) else "[embedded value]"
    )
    return ret


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



import re
import json

from .config import train_file_path, clf_json_file_path


def intent_to_label(intent):
    intent = intent.replace("_", "-")
    intent = intent.replace(" ", "-")
    return f"__label__{intent}"


def label_to_intent(label):
    label = label.replace("__label__", "")
    return label.replace("-", " ")


def prep_label(intent):
    return intent_to_label(intent)


def prep_sentence(sentence):
    processed = sentence.lower()
    processed = re.sub(r"-", " ", processed)
    processed = re.sub(r"[^a-zA-Z0-9 ]", "", processed)
    # processed = re.sub(r'[0-9]', '[D]', processed)
    return processed


def json_to_train():
    result = []
    with open(clf_json_file_path, "r", encoding="utf-8") as clf_file:
        clf_data = json.load(clf_file)
    for intent, data in clf_data.items():
        label = prep_label(intent)
        for sentence in data:
            sentence = prep_sentence(sentence)
            result.append(f"{label} {sentence}")
    with open(train_file_path, "w", encoding="utf-8") as of:
        of.write("\n".join(result))

    print(f"Wrote {len(result)} sentences to {train_file_path}")


if __name__ == "__main__":
    json_to_train()



import numpy as np
import tensorflow_hub as hub
import tensorflow as tf
import tensorflow_text  # noqa
from jaseci.actions.live_actions import jaseci_action
from typing import Union


module = hub.load(
    "https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/3"
)


@jaseci_action(act_group=["use"], aliases=["enc_question"], allow_remote=True)
def question_encode(question: Union[str, list]):
    if isinstance(question, list):
        return (
            module.signatures["question_encoder"](tf.constant(question))["outputs"]
            .numpy()
            .tolist()
        )
    elif isinstance(question, str):
        return (
            module.signatures["question_encoder"](tf.constant([question]))["outputs"]
            .numpy()
            .tolist()
        )


@jaseci_action(act_group=["use"], aliases=["enc_answer"], allow_remote=True)
def answer_encode(answer: Union[str, list], context: Union[str, list] = None):
    if context is None:
        context = answer
    if isinstance(answer, list):
        return (
            module.signatures["response_encoder"](
                input=tf.constant(answer), context=tf.constant(context)
            )["outputs"]
            .numpy()
            .tolist()
        )
    elif isinstance(answer, str):
        return (
            module.signatures["response_encoder"](
                input=tf.constant([answer]), context=tf.constant([context])
            )["outputs"]
            .numpy()
            .tolist()
        )


@jaseci_action(act_group=["use"], allow_remote=True)
def cos_sim_score(q_emb: list, a_emb: list):
    norm = np.linalg.norm
    return np.dot(q_emb, a_emb) / (norm(q_emb) * norm(a_emb))


@jaseci_action(act_group=["use"], aliases=["qa_score"], allow_remote=True)
def dist_score(q_emb: list, a_emb: list):
    return np.inner(q_emb, a_emb).tolist()


@jaseci_action(act_group=["use"], allow_remote=True)
def question_similarity(text1: str, text2: str):
    enc_a = np.squeeze(np.asarray(question_encode(text1)))
    enc_b = np.squeeze(np.asarray(question_encode(text2)))
    return cos_sim_score(list(enc_a), list(enc_b))


@jaseci_action(act_group=["use"], allow_remote=True)
def question_classify(text: str, classes: list):
    text_emb = np.squeeze(np.asarray(question_encode(text)))
    ret = {"match": "", "match_idx": -1, "scores": []}
    for i in classes:
        i_emb = np.squeeze(np.asarray(question_encode(i))) if isinstance(i, str) else i
        ret["scores"].append(cos_sim_score(text_emb, i_emb))
    top_hit = ret["scores"].index(max(ret["scores"]))
    ret["match_idx"] = top_hit
    ret["match"] = (
        classes[top_hit] if isinstance(classes[top_hit], str) else "[embedded value]"
    )
    return ret


@jaseci_action(act_group=["use"], allow_remote=True)
def answer_similarity(text1: str, text2: str):
    enc_a = np.squeeze(np.asarray(answer_encode(text1)))
    enc_b = np.squeeze(np.asarray(answer_encode(text2)))
    return cos_sim_score(list(enc_a), list(enc_b))


@jaseci_action(act_group=["use"], allow_remote=True)
def answer_classify(text: str, classes: list):
    text_emb = np.squeeze(np.asarray(answer_encode(text)))
    ret = {"match": "", "match_idx": -1, "scores": []}
    for i in classes:
        i_emb = np.squeeze(np.asarray(answer_encode(i))) if isinstance(i, str) else i
        ret["scores"].append(cos_sim_score(text_emb, i_emb))
    top_hit = ret["scores"].index(max(ret["scores"]))
    ret["match_idx"] = top_hit
    ret["match"] = (
        classes[top_hit] if isinstance(classes[top_hit], str) else "[embedded value]"
    )
    return ret


@jaseci_action(act_group=["use"], allow_remote=True)
def qa_similarity(text1: str, text2: str):
    enc_a = np.squeeze(np.asarray(question_encode(text1)))
    enc_b = np.squeeze(np.asarray(answer_encode(text2)))
    return cos_sim_score(list(enc_a), list(enc_b))


@jaseci_action(act_group=["use"], allow_remote=True)
def qa_classify(text: str, classes: list):
    text_emb = np.squeeze(np.asarray(question_encode(text)))
    ret = {"match": "", "match_idx": -1, "scores": []}
    for i in classes:
        i_emb = np.squeeze(np.asarray(answer_encode(i))) if isinstance(i, str) else i
        ret["scores"].append(cos_sim_score(text_emb, i_emb))
    top_hit = ret["scores"].index(max(ret["scores"]))
    ret["match_idx"] = top_hit
    ret["match"] = (
        classes[top_hit] if isinstance(classes[top_hit], str) else "[embedded value]"
    )
    return ret


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



import logging
from dataclasses import dataclass, fields
from typing import Tuple, Iterable, Optional, Union, Dict, List, Set, Callable
import torch
from torch import LongTensor
from torch.nn.utils.rnn import pad_sequence
from transformers.tokenization_utils_base import EncodingFast
from .utils import pad_images, invert


logger = logging.getLogger(__name__)


@dataclass
class TupleLike:
    def __iter__(self):  # so it behaves like a NamedTuple
        return iter(self.as_tuple())

    def as_tuple(self):
        return tuple(self.__getattribute__(field.name) for field in fields(self))


@dataclass(repr=True, eq=True, order=True, frozen=True)
class TypedSpan:
    start: int
    end: int
    type: str

    def __iter__(self):  # so it behaves like a NamedTuple
        return iter(self.as_tuple())

    def as_tuple(self):
        return tuple(self.__getattribute__(field.name) for field in fields(self))


@dataclass
class Example(TupleLike):
    text_id: int
    example_start: int
    input_ids: LongTensor  # shape: (LENGTH)
    start_offset: LongTensor
    end_offset: LongTensor
    target_label_ids: Optional[LongTensor]  # shape: (LENGTH, MAX_ENTITY_LENGTH)


@dataclass
class BatchedExamples(TupleLike):
    text_ids: Tuple[int, ...]
    example_starts: Tuple[int, ...]
    input_ids: LongTensor  # shape: (BATCH_SIZE, LENGTH)
    start_offset: LongTensor
    end_offset: LongTensor


def collate_examples(
    examples: Iterable[Example],
    *,
    padding_token_id: int,
    pad_length: Optional[int] = None,
    return_batch_examples: bool = True,
) -> Dict[str, Union[BatchedExamples, Optional[LongTensor]]]:
    all_text_ids: List[int] = []
    all_example_starts: List[int] = []
    all_input_ids: List[LongTensor] = []
    all_start_offsets: List[LongTensor] = []
    all_end_offsets: List[LongTensor] = []
    target_label_ids: Optional[List[LongTensor]] = None

    no_target_label_ids: Optional[bool] = None

    for example in examples:
        all_text_ids.append(example.text_id)
        all_example_starts.append(example.example_start)
        all_input_ids.append(example.input_ids)
        all_start_offsets.append(example.start_offset)
        all_end_offsets.append(example.end_offset)

        if no_target_label_ids is None:
            no_target_label_ids = example.target_label_ids is None
            if not no_target_label_ids:
                target_label_ids: List[LongTensor] = []

        if (example.target_label_ids is None) != no_target_label_ids:
            raise RuntimeError("Inconsistent examples at collate_examples!")

        if example.target_label_ids is not None:
            target_label_ids.append(example.target_label_ids)

    res = {
        "input_ids": pad_sequence(
            all_input_ids, batch_first=True, padding_value=padding_token_id
        ).long(),
        "labels": pad_images(
            target_label_ids, padding_value=-100, padding_length=(pad_length, None)
        )
        if not no_target_label_ids
        else None,
    }

    if return_batch_examples:
        res["examples"] = BatchedExamples(
            tuple(all_text_ids),
            tuple(all_example_starts),
            pad_sequence(
                all_input_ids, batch_first=True, padding_value=padding_token_id
            ).long(),
            pad_sequence(
                all_start_offsets, batch_first=True, padding_value=-100
            ).long(),
            pad_sequence(all_end_offsets, batch_first=True, padding_value=-100).long(),
        )
    return res


def batch_examples(
    examples: Iterable[Example],
    *,
    collate_fn: Callable[
        [Iterable[Example]], Dict[str, Union[BatchedExamples, LongTensor]]
    ],
    batch_size: int = 1,
) -> Iterable[Dict[str, Union[BatchedExamples, LongTensor]]]:
    """Groups examples into batches."""
    curr_batch = []
    for example in examples:
        if len(curr_batch) == batch_size:
            yield collate_fn(curr_batch)
            curr_batch = []

        curr_batch.append(example)

    if len(curr_batch):
        yield collate_fn(curr_batch)


def strided_split(
    text_id: int,
    encoding: EncodingFast,
    entities: Optional[Set[TypedSpan]] = None,
    *,
    category_mapping: Dict[str, int],
    no_entity_category: str,
    stride: int,
    max_sequence_length: int,
    max_entity_length: int,
    cls_token_id: int,
) -> Iterable[Example]:
    assert 0 < stride <= max_sequence_length

    sequence_length = len(encoding.ids)
    offset = torch.tensor(encoding.offsets, dtype=torch.long)

    target_label_ids: Optional[LongTensor] = None
    if entities is not None:
        token_start_mapping = {}
        token_end_mapping = {}
        for token_idx, (token_start, token_end) in enumerate(encoding.offsets):
            token_start_mapping[token_start] = token_idx
            token_end_mapping[token_end] = token_idx

        no_entity_category_id = category_mapping[no_entity_category]
        category_id_mapping = invert(category_mapping)

        text_length = len(encoding.ids)
        target_label_ids = torch.full(
            (text_length, max_entity_length),
            fill_value=no_entity_category_id,
            dtype=torch.long,
        ).long()
        for start, end, category in entities:
            try:
                token_start = token_start_mapping[start]
            except KeyError:
                if start + 1 in token_start_mapping:
                    logger.warning(f"changing {start} to {start + 1}")
                    token_start = token_start_mapping[start + 1]
                elif start - 1 in token_start_mapping:
                    logger.warning(f"changing {start} to {start - 1}")
                    token_start = token_start_mapping[start - 1]
                else:
                    logger.warning(
                        f"Skipped entity {category} at ({start} {end}) -- weird span"
                    )
                    continue

            try:
                token_end = token_end_mapping[end]
            except KeyError:  # for some reason some ends are shifted by one
                if end + 1 in token_end_mapping:
                    logger.warning(f"changing {end} to {end + 1}")
                    token_end = token_end_mapping[end + 1]
                elif end - 1 in token_end_mapping:
                    logger.warning(f"changing {end} to {end - 1}")
                    token_end = token_end_mapping[end - 1]
                else:
                    logger.warning(
                        f"Skipped entity {category} at ({start} {end}) -- weird span"
                    )
                    continue

            token_length = token_end - token_start
            if token_length >= max_entity_length:
                logger.warning(
                    f"Skipped entity {category} at ({start} {end}) -- too long"
                )
                continue

            prev_label = target_label_ids[token_start][token_length].item()

            if prev_label != no_entity_category_id:
                from_category = category_id_mapping[prev_label]
                logger.warning(
                    f"""rewriting entity of category
                     {from_category} with {category} at ({start} {end}) span"""
                )

            target_label_ids[token_start][token_length] = category_mapping[category]

    # split encoding into max_length-token chunks

    chunk_start = 0
    chunk_end = 0
    while chunk_start < sequence_length and chunk_end != sequence_length:
        chunk_end = min(
            chunk_start + max_sequence_length - 1, sequence_length
        )  # -1 for cls token
        ex = Example(
            text_id,
            chunk_start,
            torch.tensor(
                [cls_token_id] + encoding.ids[chunk_start:chunk_end], dtype=torch.long
            ).long(),
            torch.tensor(
                [-100] + offset[chunk_start:chunk_end, 0].tolist(), dtype=torch.long
            ).long(),
            torch.tensor(
                [-100] + offset[chunk_start:chunk_end, 1].tolist(), dtype=torch.long
            ).long(),
            torch.tensor(
                [[-100] * max_entity_length]
                + target_label_ids[chunk_start:chunk_end].tolist(),
                dtype=torch.long,
            ).long()
            if target_label_ids is not None
            else None,
        )
        yield ex
        chunk_start += stride



from typing import Tuple, Callable, List, Set, Iterable, Dict
import json
from transformers import BatchEncoding
from ..datamodel.example import TypedSpan, Example
from ..datamodel.dataset import NERDataset


def prepare_descriptions() -> Dict[str, str]:
    with open("/home/ubuntu/binder/data/genia/descriptions.json") as f:
        return json.load(f)


def read_annotation(annotations) -> Set[TypedSpan]:
    collected_annotations: Set[TypedSpan] = set()
    for line in annotations:
        collected_annotations.add(
            TypedSpan(
                int(line["start_index"]), int(line["end_index"]), line["entity_type"]
            )
        )
    return collected_annotations


def read_data(dataset: Dict) -> Tuple[List[str], List[Set[TypedSpan]]]:
    list1, list2 = [], []
    for text, entity in zip(dataset["text"], dataset["annotations"]):
        list1.append(text)
        list2.append(read_annotation(entity))
    return list1, list2


def get_datasets(
    dataset: Dict,
    example_encoder: Callable[[List[str], List[Set[TypedSpan]]], Iterable[Example]],
) -> Tuple[NERDataset]:
    return NERDataset(example_encoder(*read_data(dataset)))


def tokenize_descriptions(
    descriptions: List[str], tokenizer, max_length
) -> BatchEncoding:
    return tokenizer(
        descriptions,
        return_tensors="pt",
        truncation=True,
        max_length=max_length,
        add_special_tokens=True,
        padding=True,
    )



import torch
from torch.utils.data import DataLoader
import os
from tqdm.autonotebook import trange
from transformers.optimization import AdamW, get_linear_schedule_with_warmup
from . import tokenizer as token_util
import math


def train_model(model, tokenizer, contexts, candidates, labels, train_config):
    context_transform = token_util.SelectionJoinTransform(
        tokenizer=tokenizer, max_len=train_config["max_contexts_length"]
    )
    candidate_transform = token_util.SelectionSequentialTransform(
        tokenizer=tokenizer, max_len=int(train_config["max_candidate_length"])
    )
    train_dataset = token_util.SelectionDataset(
        contexts=contexts,
        candidates=candidates,
        labels=labels,
        context_transform=context_transform,
        candidate_transform=candidate_transform,
    )
    train_dataloader = DataLoader(
        train_dataset,
        batch_size=train_config["train_batch_size"],
        collate_fn=train_dataset.batchify_join_str,
        shuffle=True,
    )
    t_total = (
        len(train_dataloader)
        // train_config["gradient_accumulation_steps"]
        * (max(5, train_config["num_train_epochs"]))
    )

    global_step = 0
    if not os.path.exists(train_config["basepath"]):
        os.makedirs(train_config["basepath"])
    log_wf = open(
        os.path.join(train_config["basepath"], "log.txt"), "w", encoding="utf-8"
    )  # will be used in for logging
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": train_config["weight_decay"],
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]
    optimizer = AdamW(
        optimizer_grouped_parameters,
        lr=train_config["learning_rate"],
        eps=train_config["adam_epsilon"],
    )
    warmup_steps = math.ceil(
        len(train_dataset)
        * train_config["num_train_epochs"]
        / train_config["train_batch_size"]
        * 0.1
    )
    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total
    )
    fp16 = False
    if fp16:
        try:
            from apex import amp
        except ImportError:
            raise ImportError(
                """Please install apex from https://www.github.com/nvidia/apex
                to use fp16 training"""
            )
        model, optimizer = amp.initialize(
            model, optimizer, opt_level=train_config["fp16_opt_level"]
        )
    for epoch in trange(
        train_config["num_train_epochs"], desc="Epoch", disable=False, unit="batch"
    ):
        tr_loss = 0
        nb_tr_examples, nb_tr_steps = 0, 0
        with trange(len(train_dataloader), unit="it") as bar:
            for step, batch in enumerate(train_dataloader, start=1):
                model.train()
                optimizer.zero_grad()
                batch = tuple(t.to(train_config["device"]) for t in batch)
                (
                    context_token_ids_list_batch,
                    context_input_masks_list_batch,
                    candidate_token_ids_list_batch,
                    candidate_input_masks_list_batch,
                    labels_batch,
                ) = batch

                loss = model(
                    context_token_ids_list_batch,
                    context_input_masks_list_batch,
                    candidate_token_ids_list_batch,
                    candidate_input_masks_list_batch,
                    labels_batch,
                )
                tr_loss += loss.item()
                nb_tr_examples += context_token_ids_list_batch.size(0)
                nb_tr_steps += 1
                if (step + 1) % train_config["gradient_accumulation_steps"] == 0:
                    if fp16:
                        with amp.scale_loss(loss, optimizer) as scaled_loss:
                            scaled_loss.backward()
                        torch.nn.utils.clip_grad_norm_(
                            amp.master_params(optimizer), train_config["max_grad_norm"]
                        )
                    else:
                        loss.backward()
                        torch.nn.utils.clip_grad_norm_(
                            model.parameters(), train_config["max_grad_norm"]
                        )

                    optimizer.step()
                    scheduler.step()
                    model.zero_grad()
                    global_step += 1
                    bar.update()

        print(
            f"""\n
            Epoch : {epoch+1}
            loss : {tr_loss/nb_tr_steps}
            LR : {optimizer.param_groups[0]['lr']}\n"""
        )
        log_wf.write(f"{epoch+1}\t{tr_loss/nb_tr_steps}\n")
    log_wf.close()
    return model



import os
import openai
from jaseci.actions.live_actions import jaseci_action
import traceback
from fastapi import HTTPException

DEFAULT_ARGS = {
    "model": "text-davinci-003",
    "temperature": 0,
    "max_tokens": 100,
    "top_p": 1,
    "frequency_penalty": 0.0,
    "presence_penalty": 0.0,
}


@jaseci_action(act_group=["gpt3"], allow_remote=True)
def setup(api_key: str):
    if api_key:
        openai.api_key = api_key
    elif os.environ.get("OPENAI_API_KEY", None):
        openai.api_key = os.environ.get("OPENAI_API_KEY")
    else:
        print(
            "No API key provided. Please set the OPENAI_API_KEY environment variable or pass it to the setup function."
        )


setup(api_key=os.environ.get("OPENAI_API_KEY", None))


@jaseci_action(act_group=["gpt3"], allow_remote=True)
def generate(text: str, args: dict = DEFAULT_ARGS):
    try:
        response = openai.Completion.create(prompt=text, **args)
        return response["choices"]
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from typing import Dict, List, Union

import numpy as np
from fastapi import HTTPException
from flair.data import Corpus, Sentence
from flair.datasets import FlairDatapointDataset
from flair.models import TARSClassifier
from flair.trainers import ModelTrainer
from jaseci.actions.live_actions import jaseci_action


@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def setup():
    """load the tars classifier for ZS classification"""
    global classifier
    model_name = "tars-base"
    classifier = TARSClassifier.load(model_name)
    print(f"loaded mode : [{model_name}]")


# initialize the classifier
setup()


# API for getting the cosine similarity
@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def cosine_sim(vec_a: List[float], vec_b: List[float]):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """

    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return result.astype(float)


# defining the api for ZS classification
@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def train(dataset: Union[Dict[str, str], List[Dict[str, str]]]):
    """
    API for classifying text among classes provided
    """
    global classifier
    label_type = "zs_classifier"
    task_name = "zs_train"
    try:
        if isinstance(dataset, list):
            sent_list = []
            for sent in dataset:
                text, label = list(sent.items())[0]
                sent_list.append(Sentence(text).add_label(label_type, label))
            train_data = FlairDatapointDataset(sent_list)
        else:
            text, label = list(dataset.items())[0]
            train_data = FlairDatapointDataset(
                [Sentence(text).add_label(label_type, label)]
            )
        corpus = Corpus(train=train_data, test=train_data, dev=train_data)
        classifier.add_and_switch_to_new_task(
            task_name,
            label_dictionary=corpus.make_label_dictionary(label_type=label_type),
            label_type=label_type,
        )
        trainer = ModelTrainer(classifier, corpus)
        trainer.train(
            base_path="model/",
            learning_rate=0.02,
            mini_batch_size=1,
            max_epochs=5,
            train_with_dev=True,
            train_with_test=True,
        )
        classifier = TARSClassifier.load("model/final-model.pt")
        return "Training Completed"
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(f"Exception :{e}"))


# defining the api for ZS classification
@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def classify(text: Union[str, List[str]], classes: List[str]):
    """
    API for classifying text among classes provided
    """
    response_data_format = []
    try:
        if isinstance(text, list):
            for sent_text in text:
                resp_data = {}
                sent = Sentence(sent_text)
                # predicting class for the text
                classifier.predict_zero_shot(sent, classes)
                pred_output = sent.to_dict()
                resp_data[pred_output["text"]] = pred_output["all labels"]
                response_data_format.append(resp_data)
        else:
            sent = Sentence(text)
            classifier.predict_zero_shot(sent, classes)
            pred_output = sent.to_dict()
            resp_data = {}
            resp_data[pred_output["text"]] = pred_output["all labels"]
            response_data_format.append(resp_data)
        return response_data_format
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(f"Exception :{e}"))


@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def get_embeddings(texts: Union[str, List[str]]):
    """
    API to get embbeddings for text
    """
    embedder = classifier.tars_embeddings
    if isinstance(texts, str):
        label_sentences = Sentence([texts])
        embedder.embed(label_sentences)
        return label_sentences.get_embedding().cpu().detach().numpy().tolist()
    else:
        embedding_list = []
        for text in texts:
            label_sentences = Sentence(text)
            embedder.embed(label_sentences)
            embedding_list.append(
                label_sentences.get_embedding().cpu().detach().numpy().tolist()
            )
        return embedding_list


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



import torch
from torch import Tensor, LongTensor
from torch.nn import Module

REDUCTION = {"mean": torch.nanmean, "sum": torch.nansum, "none": lambda x, dim: x}


def my_logsumexp(masked_tensor: Tensor, tensor: Tensor):
    maxes = torch.amax(masked_tensor, dim=[-2, -1], keepdim=True)
    maxes = torch.masked_fill(maxes, maxes.abs() == float("inf"), 0)

    result = torch.sum(
        torch.exp(masked_tensor - maxes), dim=[-2, -1], keepdim=True
    ) + torch.exp(
        tensor - maxes
    )  # might overflow
    return result.log().add(maxes)


class ContrastiveThresholdLoss(Module):
    def __init__(self, n_classes: int, ignore_id: int, reduction: str, beta: float):
        super(ContrastiveThresholdLoss, self).__init__()
        self._beta = beta
        self._n_classes = n_classes
        self._ignore_id = ignore_id
        self._reduce = REDUCTION[reduction]

    def forward(self, predicted_scores: Tensor, true_ids: LongTensor) -> Tensor:
        """Computes positive-unlabelled loss.

        :param predicted_scores: (B, S, N, C)
        :param true_ids:  (B, S, N)
        :return: loss with applied reduction
        """
        assert (
            true_ids[:, 0, 0] == self._ignore_id
        ).all()  # the first label should be [CLS]
        device = predicted_scores.device
        classes = torch.arange(self._n_classes, device=device)

        ignore_mask = (
            (true_ids == self._ignore_id).unsqueeze(1).repeat(1, self._n_classes, 1, 1)
        )  # (B, C, S, N)
        class_mask = classes.reshape(1, self._n_classes, 1, 1) == true_ids.unsqueeze(
            1
        )  # (B, C, S, N)

        denominator_mask = ~class_mask & ~ignore_mask  # elems for denominator
        denominator_mask[:, :, 0, 0] = True  # include cls score

        predicted_scores = predicted_scores.swapaxes(-2, -1).swapaxes(
            -3, -2
        )  # (B, S, N, C) -> (B, C, S, N)

        denominator_score = torch.clone(predicted_scores)
        denominator_score[~denominator_mask] = -torch.inf  # exp will turn this into 0
        denominator_score = my_logsumexp(
            denominator_score, predicted_scores
        )  # (B, C, S, N)
        cls_score = denominator_score[:, :, 0, 0] - predicted_scores[:, :, 0, 0]

        contrastive_scores = denominator_score - predicted_scores

        # mean over positives
        positive_scores_mask = ~ignore_mask & class_mask
        contrastive_scores[~positive_scores_mask] = torch.nan
        contrastive_losses = contrastive_scores.nanmean(dim=[-2, -1])

        # use [CLS] label as a threshold
        threshold_losses = (
            contrastive_losses * self._beta + (1 - self._beta) * cls_score
        )

        return self._reduce(threshold_losses)



import os
import torch
import shutil
from pathlib import Path
from functools import partial
from torch.nn import Linear, Embedding
from torch import Tensor, LongTensor, BoolTensor
from transformers.tokenization_utils_base import EncodingFast
from typing import List, Tuple, Union, Optional, TypeVar, Iterable, Set, Dict
from torch.nn.functional import pad
from transformers import (
    PreTrainedModel,
    AutoModel,
    AutoTokenizer,
    PreTrainedTokenizer,
    BatchEncoding,
)
from torch.nn.utils.rnn import pad_sequence
from .metric import CosSimilarity
from .loss import ContrastiveThresholdLoss
from ..datamodel.example import Example, strided_split, TypedSpan, BatchedExamples
from .base_encoder import Base_BI_enc
from ..datamodel.utils import pad_images

_Model = TypeVar("_Model", bound="BI_P_Head")


class BI_P_Head(Base_BI_enc):
    def __init__(self, param) -> None:
        super(BI_P_Head, self).__init__(param)
        global _token_tokenizer
        self._metric = CosSimilarity(scale=0.07)
        self._unk_entity_label_id = param.get("unk_entity_type_id")
        self._loss_fn_factory = partial(
            ContrastiveThresholdLoss,
            ignore_id=-100,
            reduction="mean",
            beta=param.get("loss_beta"),
        )
        self._token_encoder = self._context_encoder
        self._token_tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(
            param.get("context_bert_model")
        )
        _token_tokenizer = self._token_tokenizer
        self._descriptions_encoder = self._candidate_encoder
        self._descriptions_tokenizer: PreTrainedTokenizer = (
            AutoTokenizer.from_pretrained(param.get("candidate_bert_model"))
        )
        self._loss_fn: Optional[ContrastiveThresholdLoss] = None
        self._start_coef = param.get("start_coef")
        self._end_coef = param.get("end_coef")
        self._span_coef = param.get("span_coef")

        self._hidden_size = param.get("hidden_size")
        self._encoder_hidden = self._token_encoder.config.hidden_size
        self._max_sequence_length = min(
            param.get("max_sequence_length"), self._token_tokenizer.model_max_length
        )
        self._max_entity_length = param.get("max_entity_length")

        self._entity_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._encoded_descriptions: Optional[BatchEncoding] = None
        self._frozen_entity_representations: Optional[Tensor] = None
        self.add_descriptions(param.get("descriptions"))

        self._length_embedding = Embedding(self._max_entity_length, self._hidden_size)
        self._span_projection = Linear(
            self._encoder_hidden * 2 + self._hidden_size, self._hidden_size
        )
        self._token_start_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._token_end_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._entity_start_projection = Linear(self._encoder_hidden, self._hidden_size)
        self._entity_end_projection = Linear(self._encoder_hidden, self._hidden_size)

    def freeze_descriptions(self):
        if self._frozen_entity_representations is not None:
            return

        with torch.no_grad():
            self._frozen_entity_representations = self._descriptions_encoder(
                self._encoded_descriptions.to(self.device)
            ).last_hidden_state[
                :, 0
            ]  # TODO: split into batches

    def drop_descriptions_encoder(self):
        self._descriptions_encoder = None

    def add_descriptions(self, descriptions: List[str]) -> None:
        self._encoded_descriptions = self._descriptions_tokenizer(
            descriptions,
            return_tensors="pt",
            truncation=True,
            max_length=self._descriptions_encoder.config.max_length,
            add_special_tokens=True,
            padding=True,
        )["input_ids"]
        self._loss_fn = self._loss_fn_factory(len(descriptions))
        self._frozen_entity_representations = None

    def eval(self: _Model) -> _Model:
        super(BI_P_Head, self).eval()
        self.freeze_descriptions()
        return self

    def train(self: _Model, mode: bool = True) -> _Model:
        super(BI_P_Head, self).train(mode)
        self._frozen_entity_representations = None
        return self

    def _get_entity_representations(self) -> Tensor:
        if self._frozen_entity_representations is not None:
            return self._frozen_entity_representations.to(self.device)
        return self._descriptions_encoder(
            self._encoded_descriptions.to(self.device)
        ).last_hidden_state[
            :, 0
        ]  # TODO: split into batches

    def _get_length_representations(self, token_representations: Tensor) -> Tensor:
        batch_size, sequence_length, representation_dims = token_representations.shape

        length_embeddings = self._length_embedding(
            torch.arange(self._max_entity_length, device=self.device)
        )
        return length_embeddings.reshape(
            1, 1, self._max_entity_length, self._hidden_size
        ).repeat(batch_size, sequence_length, 1, 1)

    def _get_span_representations(
        self, token_representations: Tensor
    ) -> Tuple[Tensor, BoolTensor]:
        batch_size, sequence_length, representation_dims = token_representations.shape

        padding_masks: List[BoolTensor] = []
        span_end_representations: List[Tensor] = []

        for shift in range(
            self._max_entity_length
        ):  # self._max_entity_length ~ 20-30, so it is fine to not vectorize this
            span_end_representations.append(
                torch.roll(token_representations, -shift, 1).unsqueeze(-2)
            )

            padding_mask = torch.ones(
                (batch_size, sequence_length), dtype=torch.bool, device=self.device
            )
            padding_mask[:, -shift:] = False
            padding_masks.append(padding_mask.unsqueeze(-1))

        padding_mask = torch.concat(padding_masks, dim=-1).bool()
        span_end_representation = torch.concat(span_end_representations, dim=-2)
        span_start_representation = token_representations.unsqueeze(-2).repeat(
            1, 1, self._max_entity_length, 1
        )

        span_length_embedding = self._get_length_representations(token_representations)

        span_representation = torch.concat(
            [span_start_representation, span_end_representation, span_length_embedding],
            dim=-1,
        )
        return self._span_projection(span_representation), padding_mask

    def _compute_sim_scores(
        self,
        representations: Tensor,
        entity_representations: Tensor,
        *,
        span_level: bool = True,
    ) -> Tensor:
        n_classes = len(self._encoded_descriptions)

        representations = representations.unsqueeze(
            -2
        )  # (B, S, M, 1, H) or (B, S, 1, H)
        new_shape = (
            (1, 1, 1, n_classes, self._hidden_size)
            if span_level
            else (1, 1, n_classes, self._hidden_size)
        )
        entity_representations = entity_representations.reshape(*new_shape)
        return self._metric(representations, entity_representations)

    def forward(
        self, input_ids: LongTensor, labels: Optional[LongTensor] = None, **_
    ) -> Union[LongTensor, Tuple[Tensor, LongTensor]]:
        """Predicts entity type ids. If true label ids are given,
        calculates loss as well."""
        _, sequence_length = input_ids.shape
        input_ids = pad(
            input_ids,
            [0, self._max_sequence_length - sequence_length],
            value=self._token_tokenizer.pad_token_id,
        )

        token_representations = self._token_encoder(
            input_ids=input_ids.to(self.device)
        )[
            "last_hidden_state"
        ]  # (B, S, E)
        entity_representations = self._get_entity_representations()  # (C, E)

        return entity_representations, token_representations

    def get_scores(self, token_representations, entity_representations, labels=None):
        span_representations, span_padding = self._get_span_representations(
            token_representations
        )
        span_scores = self._compute_sim_scores(
            span_representations, self._entity_projection(entity_representations)
        )
        batch_size, _, _, n_classes = span_scores.shape

        entity_threshold = span_scores[:, 0, 0].reshape(
            batch_size, 1, 1, n_classes
        )  # is a sim score with [CLS] token and entities
        mask = span_scores > entity_threshold
        scores_mask = torch.full_like(span_scores, fill_value=-torch.inf)
        scores_mask[mask] = 0  # set to -inf scores that did not pass the threshold

        values, predictions = torch.max(span_scores + scores_mask, dim=-1)
        predictions[values == -torch.inf] = self._unk_entity_label_id

        if labels is None:
            return predictions

        # (C, H)
        entity_start_representations = self._entity_start_projection(
            entity_representations
        )
        entity_end_representations = self._entity_end_projection(entity_representations)

        # (B, S, H)
        token_start_representations = self._token_start_projection(
            token_representations
        )
        token_end_representations = self._token_start_projection(token_representations)
        # labels = labels.to(self.device)?
        start_scores = self._compute_sim_scores(
            token_start_representations, entity_start_representations, span_level=False
        )
        end_scores = self._compute_sim_scores(
            token_end_representations, entity_end_representations, span_level=False
        )
        return (span_scores, start_scores, end_scores)

    def save(self, save_path):
        super().save(save_path)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        cand_bert_path = os.path.join(save_path + "/cand_bert")
        cont_bert_path = os.path.join(save_path + "/cont_bert")
        if not os.path.exists(cand_bert_path):
            os.makedirs(cand_bert_path)
        if not os.path.exists(cont_bert_path):
            os.makedirs(cont_bert_path)
        self._context_encoder.save_pretrained(cont_bert_path)
        self._context_encoder.save_pretrained(cand_bert_path)
        self._descriptions_tokenizer.save_vocabulary(cand_bert_path)
        self._token_tokenizer.save_vocabulary(cont_bert_path)
        print(f"Saving non-shared model to : {save_path}")
        shutil.copyfile(
            os.path.join(
                Path(os.path.dirname(__file__)).parent, "config/t_config.json"
            ),
            os.path.join(save_path, "t_config.json"),
        )
        shutil.copyfile(
            os.path.join(
                Path(os.path.dirname(__file__)).parent, "config/m_config.json"
            ),
            os.path.join(save_path, "m_config.json"),
        )
        return f"[Saved model at] : {save_path}"

    def load(self, load_path) -> _Model:
        super().load(load_path)
        cand_bert_path = os.path.join(load_path, "cand_bert")
        cont_bert_path = os.path.join(load_path, "cont_bert")
        print(f"Loading non-shared model from : {load_path}")
        self._token_encoder: PreTrainedModel = AutoModel.from_pretrained(
            cont_bert_path, local_files_only=True
        )
        self._descriptions_encoder: PreTrainedModel = AutoModel.from_pretrained(
            cand_bert_path, local_files_only=True
        )
        self._token_tokenizer = AutoTokenizer.from_pretrained(
            cont_bert_path, local_files_only=True
        )
        self._descriptions_tokenizer = AutoTokenizer.from_pretrained(
            cand_bert_path, local_files_only=True
        )
        return self


def prepare_inputs(
    texts: List[str],
    entities: List[Optional[Set[TypedSpan]]],
    *,
    stride: float = 1 / 8,
    category_mapping: Dict[str, int],
    no_entity_category: str,
    text_lengths: Optional[List] = None,
    _max_sequence_length=128,
    _max_entity_length=30,
) -> Iterable[Example]:
    encodings: Optional[List[EncodingFast]] = _token_tokenizer(
        texts,
        truncation=False,
        add_special_tokens=False,
        return_offsets_mapping=True,
    ).encodings
    if encodings is None:
        raise ValueError(
            f"Tokenizer {_token_tokenizer} is not fast! Use fast tokenizer!"
        )

    if text_lengths is not None:
        for i in range(len(texts)):
            text_lengths[i] = len(encodings[i].ids)

    for text_idx, (encoding, entities) in enumerate(zip(encodings, entities)):
        yield from strided_split(
            text_idx,
            encoding,
            entities,
            stride=int(_max_sequence_length * stride),
            max_sequence_length=_max_sequence_length,
            max_entity_length=_max_entity_length,
            category_mapping=category_mapping,
            no_entity_category=no_entity_category,
            cls_token_id=_token_tokenizer.cls_token_id,
        )


def collate_examples(
    examples: Iterable[Example],
    *,
    padding_token_id: int,
    pad_length: Optional[int] = None,
    return_batch_examples: bool = True,
) -> Dict[str, Union[BatchedExamples, Optional[LongTensor]]]:
    all_text_ids: List[int] = []
    all_example_starts: List[int] = []
    all_input_ids: List[LongTensor] = []
    all_start_offsets: List[LongTensor] = []
    all_end_offsets: List[LongTensor] = []
    target_label_ids: Optional[List[LongTensor]] = None

    no_target_label_ids: Optional[bool] = None

    for example in examples:
        all_text_ids.append(example.text_id)
        all_example_starts.append(example.example_start)
        all_input_ids.append(example.input_ids)
        all_start_offsets.append(example.start_offset)
        all_end_offsets.append(example.end_offset)

        if no_target_label_ids is None:
            no_target_label_ids = example.target_label_ids is None
            if not no_target_label_ids:
                target_label_ids: List[LongTensor] = []

        if (example.target_label_ids is None) != no_target_label_ids:
            raise RuntimeError("Inconsistent examples at collate_examples!")

        if example.target_label_ids is not None:
            target_label_ids.append(example.target_label_ids)

    res = {
        "input_ids": pad_sequence(
            all_input_ids, batch_first=True, padding_value=padding_token_id
        ).long(),
        "labels": pad_images(
            target_label_ids, padding_value=-100, padding_length=(pad_length, None)
        )
        if not no_target_label_ids
        else None,
    }

    if return_batch_examples:
        res["examples"] = BatchedExamples(
            tuple(all_text_ids),
            tuple(all_example_starts),
            pad_sequence(
                all_input_ids, batch_first=True, padding_value=padding_token_id
            ).long(),
            pad_sequence(
                all_start_offsets, batch_first=True, padding_value=-100
            ).long(),
            pad_sequence(all_end_offsets, batch_first=True, padding_value=-100).long(),
        )
        return res
    return res["input_ids"], res["labels"]


def ph_init(model_args: Dict = None):
    global ph_model

    if model_args["descriptions"] is None:
        model_args["descriptions"] = ["PER", "ORG", "LOC", "MISC"]
    ph_model = BI_P_Head(model_args)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    ph_model.to(device)


def get_embeddings(data):
    return ph_model(data)


def get_scores(token_emb, ent_emb, mode):
    return ph_model.get_scores(token_emb, ent_emb, mode)



from random import random
from typing import List, Optional, Dict
from fastapi import HTTPException
from flair.data import Corpus
from flair.datasets import ColumnCorpus
from flair.models import TARSTagger, SequenceTagger
from flair.embeddings import WordEmbeddings, StackedEmbeddings, FlairEmbeddings
from flair.data import Sentence
from flair.trainers import ModelTrainer
import pandas as pd
from .entity_utils import create_data, create_data_new
import configparser
from jaseci.actions.live_actions import jaseci_action
import torch
import os
from pathlib import Path
from datetime import datetime
import warnings

warnings.filterwarnings("ignore")

config = configparser.ConfigParser()


# 1. initialize each embedding for LSTM
embedding_types = [
    # GloVe embeddings
    WordEmbeddings("glove"),
    # contextual string embeddings, forward
    FlairEmbeddings("news-forward"),
    # contextual string embeddings, backward
    FlairEmbeddings("news-backward"),
]

# embedding stack consists of Flair and GloVe embeddings
embeddings = StackedEmbeddings(embeddings=embedding_types)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def setup(reload=False):
    """create a tagger as per the model provided through config"""

    global tagger, NER_MODEL_NAME, NER_LABEL_TYPE, MODEL_TYPE, config
    if reload is False:
        config.read(os.path.join(os.path.dirname(__file__), "config.cfg"))
    NER_MODEL_NAME = config["TAGGER_MODEL"]["NER_MODEL"]
    NER_LABEL_TYPE = config["LABEL_TYPE"]["NER"]
    MODEL_TYPE = config["TAGGER_MODEL"]["MODEL_TYPE"]
    if MODEL_TYPE.lower() == "trfmodel" and NER_MODEL_NAME.lower() != "none":
        tagger = TARSTagger(embeddings=NER_MODEL_NAME)
    elif NER_MODEL_NAME.lower() != "none" and MODEL_TYPE.lower() in ["lstm", "gru"]:
        tagger = SequenceTagger.load(NER_MODEL_NAME)
    elif MODEL_TYPE.lower() in "tars" and NER_MODEL_NAME.lower() != "none":
        tagger = TARSTagger.load(NER_MODEL_NAME)
    print(f"loaded mode : [{NER_MODEL_NAME}]")


# initialize the tagger
setup()


def train_entity(train_params: dict):
    """
    funtion for training the model
    """
    global tagger, NER_MODEL_NAME, MODEL_TYPE
    # define columns
    columns = {0: "text", 1: "ner"}
    # directory where the data resides
    data_folder = "train"
    # initializing the corpus
    corpus: Corpus = ColumnCorpus(data_folder, columns, train_file="train.txt")
    # make tag dictionary from the corpus
    tag_dictionary = corpus.make_tag_dictionary(tag_type=NER_LABEL_TYPE)

    # make the model aware of the desired set of labels from the new corpus
    try:
        if MODEL_TYPE.lower() in ["trfmodel", "tars"]:
            # initialize Tars tagger with a new task
            val = random()
            tagger.add_and_switch_to_new_task(
                "ner_train" + str(val),
                label_dictionary=tag_dictionary,
                label_type=NER_LABEL_TYPE,
            )
        elif tagger is None and MODEL_TYPE.lower() in ["lstm", "gru"]:
            # initialize sequence tagger
            tagger = SequenceTagger(
                hidden_size=256,
                embeddings=embeddings,
                tag_dictionary=tag_dictionary,
                tag_type=NER_LABEL_TYPE,
                rnn_type=MODEL_TYPE,
            )
    except Exception as e:
        raise HTTPException(status_code=404, detail=str(f"Model load error : {e}"))
    #  initialize the Sequence trainer with your corpus
    trainer = ModelTrainer(tagger, corpus)
    # train model
    trainer.train(
        base_path=f"train/{NER_MODEL_NAME}",  # path to store model
        learning_rate=train_params["LR"],
        mini_batch_size=train_params["batch_size"],
        max_epochs=train_params["num_epoch"],
        train_with_test=True,
        train_with_dev=True,
    )
    #  Load trained Sequence Tagger model
    tagger = tagger.load(f"train/{NER_MODEL_NAME}/final-model.pt")
    torch.cuda.empty_cache()
    print("model training and loading completed")


# creating train val and test dataset
def create_train_data(dataset, fname):
    data = pd.DataFrame(columns=["text", "annotation"])
    for t_data in dataset:
        tag = []
        for ent in t_data["entities"]:
            if ent["entity_value"] and ent["entity_type"]:
                tag.append(
                    (
                        ent["entity_value"],
                        ent["entity_type"],
                        ent["start_index"],
                        ent["end_index"],
                    )
                )
            else:
                raise HTTPException(
                    status_code=404, detail=str("Entity Data missing in request")
                )
        data = data.append(
            {"text": t_data["context"], "annotation": tag}, ignore_index=True
        )
    # creating training data
    try:
        completed = create_data(data, fname)
    except Exception as e:
        completed = create_data_new(data, fname)
        print(f"Exception  : {e}")
    return completed


def train_and_val_entity(train_params: dict):
    """
    funtion for training the model
    """
    global tagger, NER_MODEL_NAME, MODEL_TYPE
    train_with_test = False
    train_with_dev = False
    # define columns
    columns = {0: "text", 1: "ner"}
    # directory where the data resides
    data_folder = "train"
    # initializing the corpus
    if (
        os.path.isfile("train/val.txt")
        and os.path.isfile("train/test.txt")
        and os.path.isfile("train/train.txt")
    ):
        corpus: Corpus = ColumnCorpus(
            data_folder,
            columns,
            train_file="train.txt",
            dev_file="val.txt",
            test_file="test.txt",
        )
        train_with_test = False
        train_with_dev = False
    elif os.path.isfile("train/test.txt") and os.path.isfile("train/train.txt"):
        corpus: Corpus = ColumnCorpus(
            data_folder,
            columns,
            train_file="train.txt",
            test_file="test.txt",
        )
        train_with_test = False
        train_with_dev = True
    elif os.path.isfile("train/val.txt") and os.path.isfile("train/train.txt"):
        corpus: Corpus = ColumnCorpus(
            data_folder,
            columns,
            train_file="train.txt",
            dev_file="val.txt",
        )
        train_with_test = True
        train_with_dev = False
    elif os.path.isfile("train/train.txt"):
        corpus: Corpus = ColumnCorpus(data_folder, columns, train_file="train.txt")
        train_with_test = True
        train_with_dev = True
    else:
        print("cannot find the training file ")
        return
    cp_path = Path("train/tars-ner/checkpoint.pt")
    if (cp_path).exists():
        print("in exists : ", cp_path)
        tagger = TARSTagger.load(cp_path)
        # make tag dictionary from the corpus
        tag_dictionary = corpus.make_label_dictionary(label_type=NER_LABEL_TYPE)
        # initialize sequence tagger
        try:
            tagger.add_and_switch_to_new_task(
                "ner_train_nerd",
                label_dictionary=tag_dictionary,
                label_type=NER_LABEL_TYPE,
            )
        except Exception as e:
            raise HTTPException(status_code=404, detail=str(f"Model load error : {e}"))
        #  initialize the Sequence trainer with your corpus
        trainer = ModelTrainer(tagger, corpus)
        # train model
        trainer.resume(
            model=tagger,  # path to store model
            learning_rate=train_params["LR"],
            mini_batch_size=train_params["batch_size"],
            max_epochs=train_params["num_epoch"],
            train_with_test=train_with_test,
            train_with_dev=train_with_dev,
            embeddings_storage_mode="cuda",
            checkpoint=True,
        )
    else:
        # make tag dictionary from the corpus
        tag_dictionary = corpus.make_label_dictionary(label_type=NER_LABEL_TYPE)
        try:
            tagger.add_and_switch_to_new_task(
                "ner_train_nerd",
                label_dictionary=tag_dictionary,
                label_type=NER_LABEL_TYPE,
            )
        except Exception as e:
            raise HTTPException(status_code=404, detail=str(f"Model load error : {e}"))
        #  initialize the Sequence trainer with your corpus
        trainer = ModelTrainer(tagger, corpus)
        # train model
        trainer.train(
            base_path=f"train/{NER_MODEL_NAME}",  # path to store model
            learning_rate=train_params["LR"],
            mini_batch_size=train_params["batch_size"],
            max_epochs=train_params["num_epoch"],
            train_with_test=train_with_dev,
            train_with_dev=train_with_dev,
            checkpoint=True,
        )
    #  Load trained Sequence Tagger model
    tagger = tagger.load(f"train/{NER_MODEL_NAME}/final-model.pt")
    torch.cuda.empty_cache()
    print("model training and loading completed")


# defining the api for entity detection
@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def entity_detection(text: str, ner_labels: Optional[List] = ["PREDEFINED"]):
    """
    API for detectiing provided entity in text
    """
    global tagger, MODEL_TYPE
    if tagger is not None:
        if text:
            if ner_labels:
                if MODEL_TYPE.lower() in ["trfmodel", "tars"]:
                    val = random()
                    tagger.add_and_switch_to_new_task(
                        "Entity Detection Task" + str(val),
                        ner_labels,
                        label_type=NER_LABEL_TYPE,
                    )
                sentence = Sentence(text)
                # predicting entities in the text
                tagger.predict(sentence)
                tagged_sentence = sentence.to_dict(NER_LABEL_TYPE)
                response_data_format = {"entities": []}
                response_data_format["text"] = tagged_sentence["text"]
                for entity, entity_details in zip(
                    sentence.get_spans(NER_LABEL_TYPE),
                    tagged_sentence[NER_LABEL_TYPE],
                ):
                    temp_dict = {}
                    temp_dict["entity_text"] = entity.text
                    temp_dict["entity_value"] = entity_details["value"]
                    temp_dict["conf_score"] = entity_details["confidence"]
                    temp_dict["start_pos"] = entity.start_position
                    temp_dict["end_pos"] = entity.end_position
                    response_data_format["entities"].append(temp_dict)
                return response_data_format
            else:
                raise HTTPException(
                    status_code=404,
                    detail=str("NER Labels are missing in request data"),
                )
        else:
            raise HTTPException(
                status_code=404, detail=str("Text data is missing in request data")
            )
    else:
        raise HTTPException(
            status_code=404, detail=str("Please train the Model before Using")
        )


@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def train(
    train_data: List[dict] = [],
    val_data: Optional[List[dict]] = [],
    test_data: Optional[List[dict]] = [],
    train_params: Dict = {"num_epoch": 10, "batch_size": 8, "LR": 0.02},
):
    """
    API for training the model
    """
    time1 = datetime.now()
    if len(val_data) != 0:
        create_train_data(val_data, "val")

    if len(test_data) != 0:
        create_train_data(test_data, "test")

    if len(train_data) != 0:
        completed = create_train_data(train_data, "train")
        if completed is True:
            train_and_val_entity(train_params)
            total_time = datetime.now() - time1
            print(f"total time taken to complete Training : {total_time}")
            return "Model Training is Completed"
        else:
            raise HTTPException(
                status_code=500,
                detail=str("Issue encountered during train data creation"),
            )
    else:
        raise HTTPException(
            status_code=404, detail=str("Need Data for Text and Entity")
        )


@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def save_model(model_path: str):
    """
    saves the model to the provided model_path
    """
    global tagger
    if tagger is not None:
        try:
            if not model_path.isalnum():
                raise HTTPException(
                    status_code=400,
                    detail="Invalid model name,Only alphanumeric chars allowed",
                )
            if type(model_path) is str:
                model_path = Path(model_path)
            model_path.mkdir(exist_ok=True, parents=True)
            tagger.save(model_path / "final-model.pt", checkpoint=False)
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
    else:
        raise HTTPException(
            status_code=404, detail=str("Please Initailize the model before saving")
        )


@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def load_model(model_path="default_path", default=False):
    """
    loads the model from the provided model_path
    """
    global tagger
    try:
        if default is True:
            setup()
            return f"[deafaul model loaded model from] : {config['TAGGER_MODEL']['NER_MODEL']}"  # noqa
        if type(model_path) is str:
            model_path = Path(model_path)
        if (model_path / "final-model.pt").exists():
            tagger.load_state_dict(
                tagger.load(model_path / "final-model.pt").state_dict()
            )
        tagger.to(device)
        return f"[loaded model from] : {model_path}"
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ent_ext"], allow_remote=True)
def set_config(ner_model: str = None, model_type: str = None):
    """
    Update the configuration file with new model parameters
    ner_model types:
    1. Pre-trained LSTM / GRU : ["ner", "ner-fast","ner-large"]
    2. Huggingface model : all available models
    model_type :
    1. "TRFMODEL" : for huggingface models
    2. "LSTM" or "GRU" : RNN models
    """
    global config
    config.read(os.path.join(os.path.dirname(__file__), "config.cfg"))

    if ner_model or model_type:
        config["TAGGER_MODEL"]["NER_MODEL"] = ner_model
        config["TAGGER_MODEL"]["MODEL_TYPE"] = model_type
    with open("config.cfg", "w") as configfile:
        config.write(configfile)
    try:
        setup(reload=True)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return "Config setup is complete."


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



import os
import configparser

import numpy as np

from random import sample
from typing import List

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from transformers import T5ForConditionalGeneration, T5Tokenizer, pipeline
from sentence_transformers import SentenceTransformer

config = configparser.ConfigParser()
config.read(os.path.join(os.path.dirname(__file__), "config.cfg"))


##################################################################################
#                                                                                #
#                  Model Configurations                                          #
#                                                                                #
##################################################################################

tokenizer = T5Tokenizer.from_pretrained(config["MODELS_NAMES"]["TOKENIZER"])
language_model = T5ForConditionalGeneration.from_pretrained(
    config["MODELS_NAMES"]["LANGUAGE_MODEL"]
)
headline_generator = T5ForConditionalGeneration.from_pretrained(
    config["MODELS_NAMES"]["HEADLINE_GENERATOR"]
)

sentence_embed = SentenceTransformer(config["MODELS_NAMES"]["DOCUMENT_EMBED"])

summarizer = pipeline(
    "summarization",
    model=language_model,
    tokenizer=tokenizer,
    framework="tf",
    max_length=50,
    min_length=5,
)


def c_tf_idf(documents, m, ngram_range=(1, 1)):
    """
    Calculating class based tf_idf vectors.

    Parameters:
    ------------
    documents : list of text documents.
    m : number of documents.
    ngram_range: tupple.

    Returns:

    tf_idf: calculated if_idf vectors
    count: countvectorizer object.
    """
    count = CountVectorizer(ngram_range=ngram_range, stop_words="english").fit(
        documents
    )
    t = count.transform(documents).toarray()
    w = t.sum(axis=1)
    tf = np.divide(t.T, w)
    sum_t = t.sum(axis=0)
    idf = np.log(np.divide(m, sum_t)).reshape(-1, 1)
    tf_idf = np.multiply(tf, idf)

    return tf_idf, count


def extract_top_n_words_per_topic(tf_idf, count, docs_per_topic, n=5):
    """
    Extracting top n words from each class

    Parameters:
    ------------
    tf_idf: tf-idf vectors
    count:
    docs_per_topic:
    n: number of topic expected

    Returns:
    top_n_words:  Dict,
    """
    words = count.get_feature_names_out()
    labels = list(docs_per_topic.label)
    tf_idf_transposed = tf_idf.T
    indices = tf_idf_transposed.argsort()[:, -n:]
    top_n_words = {
        label: [(words[j], tf_idf_transposed[i][j]) for j in indices[i]][::-1]
        for i, label in enumerate(labels)
    }
    return top_n_words


def generate_topic_label(documents: List[str]) -> str:
    """
    Generate summery for grouped documents in each label, and generate a headline assigined to each label.

    Parameters:
    -------------
    articles : List of Strings,

    Returns:
    -------------
    head_line: Dict, a dictio

    """

    summary_list = []
    current_token_length = 0
    max_token_length = 512

    for article in sample(documents, k=len(documents)):
        summary = summarizer(article)[0]["summary_text"]
        current_token_length += len(tokenizer.encode(summary))
        if current_token_length >= max_token_length:
            break
        summary_list.append(summary)

    encoding = tokenizer.encode(
        "headline: " + " ".join(summary_list), return_tensors="pt"
    )
    output = headline_generator.generate(encoding)
    head_line = tokenizer.decode(output[0][1:-1])

    return head_line


def mmr(input_text, top_n, n_gram_range=(1, 1), diversity=0.02):
    """
    Maximal Marginal Relevance based algorithm for extracting keywords
    """

    # Extract similarity within words, and between words and the document
    doc_embedding = sentence_embed.encode([input_text])
    count = CountVectorizer(ngram_range=n_gram_range, stop_words="english").fit(
        [input_text]
    )
    candidates = count.get_feature_names_out()
    candidate_embeddings = sentence_embed.encode(candidates)

    word_doc_similarity = cosine_similarity(candidate_embeddings, doc_embedding)
    word_similarity = cosine_similarity(candidate_embeddings)

    # Initialize candidates and already choose best keyword/keyphras
    keywords_idx = [np.argmax(word_doc_similarity)]
    candidates_idx = [i for i in range(len(candidates)) if i != keywords_idx[0]]

    for _ in range(top_n - 1):
        # Extract similarities within candidates and
        # between candidates and selected keywords/phrases
        candidate_similarities = word_doc_similarity[candidates_idx, :]
        target_similarities = np.max(
            word_similarity[candidates_idx][:, keywords_idx], axis=1
        )

        # Calculate MMR
        mmr = (
            1 - diversity
        ) * candidate_similarities - diversity * target_similarities.reshape(-1, 1)
        mmr_idx = candidates_idx[np.argmax(mmr)]

        # Update keywords & candidates
        keywords_idx.append(mmr_idx)
        candidates_idx.remove(mmr_idx)

    return [candidates[idx] for idx in keywords_idx]



from setuptools import setup, find_packages
from os.path import join

MODULES = [
    "bart_sum",
    "cl_summer",
    "ent_ext",
    "fast_enc",
    "sbert_sim",
    "t5_sum",
    "text_seg",
    "tfm_ner",
    "use_enc",
    "use_qa",
    "zs_classifier",
    "bi_enc",
    "topic_ext",
    "gpt2",
    "bi_ner",
    "gpt3",
    "sentiment",
]


def get_ver():
    with open(join("./jac_nlp", "VERSION")) as version_file:
        return version_file.read().strip()


def get_extras_requires():
    extras_requires = {"all": []}
    for module in MODULES:
        with open(join("./jac_nlp", module, "requirements.txt")) as req_file:
            extras_requires[module] = req_file.read().splitlines()
            extras_requires["all"].extend(extras_requires[module])
    return extras_requires


setup(
    name="jac_nlp",
    version=get_ver(),
    packages=find_packages(include=["jac_nlp", "jac_nlp.*"]),
    install_requires=["jaseci", "pytest>=7.0.1,<7.1", "pytest-order>=1.0.1,<1.1"],
    extras_require=get_extras_requires(),
    package_data={
        "": ["*.json", "*.cfg", "VERSION", "*.yaml", "requirements.txt", "*.jac"],
    },
    author="Jason Mars",
    author_email="jason@jaseci.org",
    url="https://github.com/Jaseci-Labs/jaseci",
)



import warnings

from jaseci.actions.live_actions import jaseci_action
from jaseci.actions.remote_actions import launch_server
from fastapi import HTTPException

from transformers import pipeline

model = pipeline(model="finiteautomata/bertweet-base-sentiment-analysis")


@jaseci_action(act_group=["sentiment"], allow_remote=True)
def predict(texts: list):
    try:
        prediction_list = model(texts)
        return prediction_list

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))



from collections import defaultdict
from copy import deepcopy
from functools import partial
from os import cpu_count
from typing import Dict, List, Set, TypeVar, Optional
import torch
from torch import LongTensor
from torch.nn import Module
from torch.nn.functional import pad
from tqdm import tqdm

from ..datamodel.example import TypedSpan, batch_examples, BatchedExamples
from ..datamodel.utils import invert
from .base_encoder import Base_BI_enc, BI_Enc_NER

torch.set_num_threads(cpu_count() // 2)


_Model = TypeVar("_Model", bound=Module)


class Bi_NER_Infer(Base_BI_enc):
    def __init__(
        self,
        span_classifier: BI_Enc_NER,
        category_mapping: Dict[str, int],
        no_entity_category: str,
        max_sequence_length: int,
        max_entity_length: int,
        model_args,
    ):
        super(Bi_NER_Infer, self).__init__(model_args)
        self._classifier = deepcopy(span_classifier)
        self._classifier.eval()
        self._classifier.drop_descriptions_encoder()
        self._classifier.train = None
        self._encoded_descriptions = self._classifier._encoded_descriptions
        self._category_mapping = deepcopy(category_mapping)
        self._category_id_mapping = invert(self._category_mapping)
        self._no_entity_category = no_entity_category
        self._max_sequence_length = max_sequence_length
        self._max_entity_length = max_entity_length
        self._text_length: List[Optional[int]] = None
        self._no_entity_category_id = None
        self._stride_length = None

    def train(self: _Model, mode: bool = True) -> _Model:
        raise NotImplementedError

    @torch.no_grad()
    def add_entity_types(self, descriptions: Dict[str, str]) -> None:
        names, texts = map(list, zip(*descriptions.items()))
        entity_representations = self._classifier._description_encoder(texts)
        self._encoded_descriptions.update(zip(names, entity_representations))

    @torch.no_grad()
    def forward(self, texts: List[str]) -> List[Set[TypedSpan]]:
        stride = 0.9
        self._stride_length = int(self._max_sequence_length * stride)

        self._text_length: List[Optional[int]] = [None] * len(texts)
        examples = list(
            self._classifier.prepare_inputs(
                texts,
                [None] * len(texts),
                category_mapping=self._category_mapping,
                no_entity_category=self._no_entity_category,
                stride=stride,
                text_lengths=self._text_length,
            )
        )

        self._no_entity_category_id = self._category_mapping[self._no_entity_category]

        predictions_collector = [defaultdict(int) for _ in texts]

        for batch in tqdm(
            batch_examples(
                examples,
                batch_size=1,
                collate_fn=partial(
                    self._classifier.collate_examples, return_batch_examples=True
                ),
            ),
            total=len(examples),
        ):
            predictions: LongTensor = self._classifier(**batch)

            batched_examples: BatchedExamples = batch["examples"]

            batch_size, length = batched_examples.start_offset.shape
            span_start = (
                pad(
                    batched_examples.start_offset,
                    [0, self._max_sequence_length - length],
                    value=-100,
                )
                .view(batch_size, self._max_sequence_length, 1)
                .repeat(1, 1, self._max_entity_length)
                .to(self.device)
            )

            end_offset = pad(
                batched_examples.end_offset,
                [0, self._max_sequence_length - length],
                value=-100,
            ).to(self.device)

            padding_masks = []
            span_end = []
            for shift in range(
                self._max_entity_length
            ):  # self._max_entity_length ~ 20-30, so it is fine to not vectorize this
                span_end.append(torch.roll(end_offset, -shift, 1).unsqueeze(-1))
                padding_mask = torch.roll(end_offset != -100, -shift, 1)
                padding_mask[:, -shift:] = False
                padding_masks.append(padding_mask.unsqueeze(-1))

            span_end = torch.concat(span_end, dim=-1)
            padding_mask = torch.concat(padding_masks, dim=-1)

            entities_mask = (
                (predictions != self._no_entity_category_id)
                & padding_mask
                & (span_end != -100)
                & (span_start != -100)
            )
            entity_token_start = (
                torch.arange(self._max_sequence_length)
                .reshape(1, self._max_sequence_length, 1)
                .repeat(batch_size, 1, self._max_entity_length)
                .to(self.device)
            )

            entity_text_ids = (
                torch.tensor(batched_examples.text_ids)
                .view(batch_size, 1, 1)
                .repeat(1, self._max_sequence_length, self._max_entity_length)
                .to(self.device)
            )

            chosen_text_ids = entity_text_ids[entities_mask]
            chosen_category_ids = predictions[entities_mask]
            chosen_span_starts = span_start[entities_mask]
            chosen_span_ends = span_end[entities_mask]
            chosen_token_starts = entity_token_start[entities_mask]
            for text_id, category_id, start, end, token_start in zip(
                chosen_text_ids,
                chosen_category_ids,
                chosen_span_starts,
                chosen_span_ends,
                chosen_token_starts,
            ):
                predictions_collector[text_id][
                    (
                        TypedSpan(
                            start.item(),
                            end.item(),
                            self._category_id_mapping[category_id.item()],
                        ),
                        token_start.item(),
                    )
                ] += 1

        all_entities = [set() for _ in texts]
        for text_id, preds in enumerate(predictions_collector):
            text_length = self._text_length[text_id]
            strided_text_length = (
                (text_length // self._stride_length)
                + (text_length % self._stride_length > 0)
            ) * self._stride_length
            for (entity, entity_token_start), count_preds in preds.items():
                # [1, 2, 3, ..., MAX, MAX, ..., MAX, MAX - 1, ..., 3, 2, 1]
                #  bin sizes are stride_length except for the last bin
                total_predictions = min(
                    (entity_token_start // self._stride_length) + 1,
                    (
                        max(strided_text_length - self._max_sequence_length, 0)
                        // self._stride_length
                    )
                    + 1,
                    ((strided_text_length - entity_token_start) // self._stride_length)
                    + 1,
                )
                if count_preds >= total_predictions // 2:
                    all_entities[text_id].add(entity)

        return all_entities



"""
This action is inactive and not maintained for produciton usage.
"""
import os
import torch
from typing import Dict, List, Union
from fastapi import HTTPException
from transformers import AutoModel, AutoConfig, AutoTokenizer
from utils.evaluate import get_embeddings, get_inference
from utils.models import PolyEncoder
import traceback
import numpy as np
from utils.train import train_model
from jaseci.actions.live_actions import jaseci_action
import random
import json
import shutil


# device = torch.device("cpu")
# uncomment this if you wish to use GPU to train
# this is commented out because this causes issues with
# unittest on machines with GPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


# funtion to set seed for the module
def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def config_setup():
    """
    Loading configurations from utils/config.cfg and
    initialize tokenizer and model
    """
    global model, tokenizer, model_config, train_config
    dirname = os.path.dirname(__file__)
    m_config_fname = os.path.join(dirname, "utils/model_config.json")
    t_config_fname = os.path.join(dirname, "utils/train_config.json")
    with open(m_config_fname, "r") as jsonfile:
        model_config = json.load(jsonfile)
    with open(t_config_fname, "r") as jsonfile:
        train_config = json.load(jsonfile)

    train_config.update({"device": device.type})
    trf_config = AutoConfig.from_pretrained(model_config["model_name"])
    tokenizer = AutoTokenizer.from_pretrained(
        model_config["model_name"], do_lower_case=True, clean_text=False
    )
    if model_config["shared"] is True:
        cont_bert = AutoModel.from_config(trf_config)
        cand_bert = cont_bert
        print("shared model created")
    else:
        cont_bert = AutoModel.from_config(trf_config)
        cand_bert = AutoModel.from_config(trf_config)
        print("non shared model created")
    model = PolyEncoder(
        config=trf_config,
        cont_bert=cont_bert,
        cand_bert=cand_bert,
        shared=model_config["shared"],
        loss_type=model_config["loss_type"],
        loss_function=model_config["loss_function"],
        poly_m=16,
    )

    model.to(train_config["device"])
    set_seed(train_config["seed"])


config_setup()


# API for getting the cosine similarity
@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def cosine_sim(vec_a: List[float], vec_b: List[float], meta):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """

    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return result.astype(float)


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def dot_prod(vec_a: List[float], vec_b: List[float]):
    """
    Caculate the dot product of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - dot product
    """
    dot_product = np.matmul(vec_a, vec_b)
    return dot_product.astype(float)


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def infer(
    contexts: Union[List[str], List[List]],
    candidates: Union[List[str], List[List]],
    context_type: str,
    candidate_type: str,
):
    """
    Take list of context, candidate and return nearest candidate to the context
    """
    model.eval()
    predicted_candidates = []
    try:
        if (context_type == "text") and (candidate_type == "text"):
            predicted_candidates.append(
                get_inference(
                    model=model,
                    contexts=contexts,
                    candidates=candidates,
                    tokenizer=tokenizer,
                    train_config=train_config,
                )
            )
            return predicted_candidates
            # con_embed = [get_context_emb(contexts)]
            # cand_embed = get_candidate_emb(candidates)
        elif (context_type == "text") and (candidate_type == "embedding"):
            con_embed = [get_context_emb(contexts)]
            cand_embed = candidates
        elif (context_type == "embedding") and (candidate_type == "text"):
            con_embed = contexts
            cand_embed = get_candidate_emb(candidates)

        elif (context_type == "embedding") and (candidate_type == "embedding"):
            con_embed = contexts
            cand_embed = candidates
        else:
            raise HTTPException(status_code=404, detail=str("input type not supported"))
        for data in con_embed:
            score_dat = []
            for lbl in cand_embed:
                if model_config["loss_type"] == "cos":
                    score_dat.append(cosine_sim(vec_a=data, vec_b=lbl))
                else:
                    score_dat.append(dot_prod(vec_a=data, vec_b=lbl))
            if candidate_type == "embedding":
                predicted_candidates.append(int(np.argmax(score_dat)))
            else:
                predicted_candidates.append(candidates[np.argmax(score_dat)])
        return predicted_candidates
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(
            status_code=404,
            detail=str(
                f"""input type can be
         'embedding' or 'text', context and
        candidate type should match the content of contexts and candidates.
        Exception : {e}"""
            ),
        )


# API for training
@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def train(contexts: List, candidates: List, labels: List[int]):
    """
    Take list of context, candidate, labels and trains the model
    """
    global model
    model.train()
    try:
        model = train_model(
            model=model,
            tokenizer=tokenizer,
            contexts=contexts,
            candidates=candidates,
            labels=labels,
            train_config=train_config,
        )
        return "Model Training is complete."
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


# API for geting Context Embedding
@jaseci_action(act_group=["poly_enc"], aliases=["encode_context"], allow_remote=True)
def get_context_emb(contexts: List):
    """
    Take list of context and returns the embeddings
    """
    model.eval()
    embedding = get_embeddings(
        model=model,
        tokenizer=tokenizer,
        text_data=contexts,
        embed_type="context",
        train_config=train_config,
    )
    return embedding


# API for geting Candidates Embedding


@jaseci_action(act_group=["poly_enc"], aliases=["encode_candidate"], allow_remote=True)
def get_candidate_emb(candidates: List):
    """
    Take list of candidates and returns the embeddings
    """
    model.eval()
    embedding = get_embeddings(
        model,
        tokenizer,
        text_data=candidates,
        embed_type="candidate",
        train_config=train_config,
    )
    return embedding


# API for setting the training and model parameters
@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def get_train_config():
    try:
        with open("utils/train_config.json", "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def set_train_config(training_parameters: Dict = None):
    global train_config
    try:
        with open("utils/train_config.json", "w+") as jsonfile:
            train_config.update(training_parameters)
            json.dump(train_config, jsonfile, indent=4)
        return "Config setup is complete."
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def get_model_config():
    try:
        with open("utils/model_config.json", "r") as jsonfile:
            data = json.load(jsonfile)
        return data
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def set_model_config(model_parameters: Dict = None):
    global model_config
    try:
        save_model(model_config["model_save_path"])
        with open("utils/model_config.json", "w+") as jsonfile:
            model_config.update(model_parameters)
            json.dump(model_config, jsonfile, indent=4)

        config_setup()
        return "Config setup is complete."

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def save_model(model_path: str):
    """
    saves the model to the provided model_path
    """
    try:
        if not model_path.replace("_", "").isalnum():
            raise HTTPException(
                status_code=400,
                detail="""
                Invalid model name. Model Name can only have Alphanumeric
                 and '_' characters.""",
            )
        if not os.path.exists(model_path):
            os.makedirs(model_path)
        if model_config["shared"] is True:
            model.cont_bert.save_pretrained(model_path)
            tokenizer.save_vocabulary(model_path)
            print(f"Saving shared model to : {model_path}")
        else:
            cand_bert_path = os.path.join(model_path + "/cand_bert")
            cont_bert_path = os.path.join(model_path + "/cont_bert")
            if not os.path.exists(cand_bert_path):
                os.makedirs(cand_bert_path)
            if not os.path.exists(cont_bert_path):
                os.makedirs(cont_bert_path)
            tokenizer.save_vocabulary(cand_bert_path)
            tokenizer.save_vocabulary(cont_bert_path)
            model.cont_bert.save_pretrained(cont_bert_path)
            model.cand_bert.save_pretrained(cand_bert_path)
            print(f"Saving non-shared model to : {model_path}")
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "utils/train_config.json"),
            os.path.join(model_path, "train_config.json"),
        )
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "utils/model_config.json"),
            os.path.join(model_path, "model_config.json"),
        )
        return f"[Saved model at] : {model_path}"
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["poly_enc"], allow_remote=True)
def load_model(model_path):
    """
    loads the model from the provided model_path
    """
    global model, tokenizer
    if not os.path.exists(model_path):
        raise HTTPException(status_code=404, detail="Model path is not available")
    try:
        with open("utils/model_config.json", "r") as jsonfile:
            model_config_data = json.load(jsonfile)
        if model_config_data["shared"] is True:
            trf_config = AutoConfig.from_pretrained(model_path, local_files_only=True)
            tokenizer = AutoTokenizer.from_pretrained(
                model_path, do_lower_case=True, clean_text=False
            )
            cont_bert = AutoModel.from_pretrained(model_path, local_files_only=True)
            cand_bert = cont_bert
            print(f"Loading shared model from : {model_path}")
        else:
            cand_bert_path = os.path.join(model_path, "cand_bert")
            cont_bert_path = os.path.join(model_path, "cont_bert")
            print(f"Loading non-shared model from : {model_path}")
            cont_bert = AutoModel.from_pretrained(cont_bert_path, local_files_only=True)
            cand_bert = AutoModel.from_pretrained(cand_bert_path, local_files_only=True)
            trf_config = AutoConfig.from_pretrained(
                cont_bert_path, local_files_only=True
            )
            tokenizer = AutoTokenizer.from_pretrained(
                cand_bert_path, do_lower_case=True, clean_text=False
            )
        model = PolyEncoder(
            config=trf_config,
            cont_bert=cont_bert,
            cand_bert=cand_bert,
            shared=model_config_data["shared"],
            loss_type=model_config["loss_type"],
            loss_function=model_config["loss_function"],
        )
        model.to(train_config["device"])
        return f"[loaded model from] : {model_path}"
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)
